unit MCoPXPTabControl;

{
     +-----------------------------------+
     \            XPTabControl           \
     \           Version 1.0.1           \
     \               2013                \
     \       http://www.dihav.com/       \
     \      vahid_you2004@yahoo.com      \
     +-----------------------------------+
}

interface

uses
  SysUtils, Classes, Windows, Messages, CommCtrl,
  {$if CompilerVersion > 22}
    Vcl.Forms, Vcl.ImgList, Vcl.Controls, Vcl.Graphics,  Vcl.ExtCtrls, Vcl.StdCtrls
  {$else}
    Forms, ImgList, Controls, Graphics,  ExtCtrls, StdCtrls
  {$ifend}
  ;

type

  EMCoPXPTabError = Exception;
  TMCoPXPTabControl = class;
  TMCoPXPTabSheet = class;

  TMCoPXPTabGradientDirection = (gdHorizontal,gdVertical);
  TMCoPXPTabButtonsPosition = (bpAbove,bpBelow,bpLeft,bpRight);
  TMCoPXPTabArrowDirection = (adUp,adDown,adLeft,adRight);
  TMCoPXPTabButtonState = (bsNormal,bsOver,bsPressed);
  TMCoPXPTabColorScheme = (csWindows,csBlue,csOlive,csSilver,csSky,csBlack,csOrange,csPurple,csGreen,csRed,csCustom);

  TMCoPXPTabOnDrawArrow = procedure(Sender:TObject;ACanvas:TCanvas;ARect:TRect;ADirection:TMCoPXPTabArrowDirection;AState:TMCoPXPTabButtonState) of object;
  TMCoPXPTabOnDrawCloseButton = procedure(Sender:TObject;ACanvas:TCanvas;ALeft,ATop:Integer;AState:TMCoPXPTabButtonState;AFocused:Boolean) of object;
  TMCoPXPTabOnMeasureCloseButton = procedure(Sender:TObject;var AWidth,AHeight:Integer) of object;
  TMCoPXPTabOnMeasureArrow = procedure(Sender:TObject;var ASize:Integer) of object;
  TMCoPXPTabOnCloseTab = procedure(Sender:TObject;ATab:TMCoPXPTabSheet;ATabIndex:Integer) of object;
  TMCoPXPTabChanging = procedure(Sender:TObject;var AllowChange:Boolean) of object;

  TMCoPXPTabSheet = class(TWinControl)
  private
    fCaption:String;
    fImgIndex:TImageIndex;
    fTabVisible:Boolean;
    fClose,fShow,fHide:TNotifyEvent;
    fTabControl:TMCoPXPTabControl;
    procedure SetCaption(V:String);
    procedure SetImgIndex(V:TImageIndex);
    procedure SetTabVisible(V:Boolean);
    function GetIndex:Integer;
    procedure SetIndex(V:Integer);
    procedure FindTabControl;
    function GetShowIndex:Integer;
    function GetControl(Index:Integer):TControl;
    function GetControlCount:Integer;
  protected
    Img:TImage;
    procedure Notification(AComponent:TComponent;Operation:TOperation); override;
    procedure AssignTo(Dest:TPersistent); override;
    procedure CMVisibleChanged(var Message:TMessage); message CM_VISIBLECHANGED;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure ShowTab;
    procedure HideTab;
    procedure Assign(Source:TPersistent); override;

    property TabControl:TMCoPXPTabControl read fTabControl write fTabControl;
    property ShowIndex:Integer read GetShowIndex;
    property ControlCount:Integer read GetControlCount;
    property Controls[Index:Integer]:TControl read GetControl;
  published
    property Caption:String read fCaption write SetCaption;
    property ImageIndex:TImageIndex read fImgIndex write SetImgIndex default -1;
    property TabVisible:Boolean read fTabVisible write SetTabVisible default True;
    property PageIndex:Integer read GetIndex write SetIndex stored False;

    property OnClose:TNotifyEvent read fClose write fClose;
    property OnShow:TNotifyEvent read fShow write fShow;
    property OnHide:TNotifyEvent read fHide write fHide;

    property Constraints;
    property Cursor;
    property DragMode;
    property Enabled;
    property Font;
    property Height stored False;
    property HelpContext;
    property HelpKeyword;
    property HelpType;
    property Hint;
    property Left stored False;
    property Name;
    property ParentBiDiMode;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Tag;
    property Top stored False;
    property Visible stored False;
    property Width stored False;

    property OnContextPopup;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDrag;
  end;

  TMCoPXPTabControl = class(TCustomControl)
  private
    fBack:{$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap;
    fBackColor1,fBackColor2:TColor;
    fBackDir:TMCoPXPTabGradientDirection;
    fButtonWidth,fButtonHeight,fLMargin,fRMargin,fTMargin,fBMargin,fScroll:Cardinal;
    fButtonPos:TMCoPXPTabButtonsPosition;
    fButtonRot,fCustomArrow,MDown,fClose,fCustomClose,fRTL,ClsFoc:Boolean;
    fBoxColor1,fBoxColor2,fArrowColor,fOverColor,fCloseColor,fTabLine:TColor;
    fActiveTab:TMCoPXPTabSheet;
    fSheetRect,CloseButtonRect:TRect;
    fTML,fTMR,fTMT,fTMB,fImgT,fImgN,fOnTab,TabDown,LX,LY,fClsT,fClsN:Integer;
    fImages:TCustomImageList;
    fDrawArrow:TMCoPXPTabOnDrawArrow;
    fDrawClose:TMCoPXPTabOnDrawCloseButton;
    fCloseSize:TMCoPXPTabOnMeasureCloseButton;
    fArrowSize:TMCoPXPTabOnMeasureArrow;
    fCloseTab:TMCoPXPTabOnCloseTab;
    fColorScheme:TMCoPXPTabColorScheme;
    fHomePage:String;
    fChanging:TMCoPXPTabChanging;
    fChange,fPaint:TNotifyEvent;
    ImageChangeLink:TChangeLink;
    ClsBtn:TButton;

    procedure Gradient(C1,C2:TColor;Ho:Boolean;var Bi:{$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap);
    function RealToByte(R:Real):Byte;
    procedure SetBackColor1(V:TColor);
    procedure SetBackColor2(V:TColor);
    procedure SetBackDir(V:TMCoPXPTabGradientDirection);
    procedure SetButtonWidth(V:Cardinal);
    procedure SetButtonHeight(V:Cardinal);
    procedure SetLMargin(V:Cardinal);
    procedure SetRMargin(V:Cardinal);
    procedure SetTMargin(V:Cardinal);
    procedure SetBMargin(V:Cardinal);
    procedure SetButtonPos(V:TMCoPXPTabButtonsPosition);
    procedure SetButtonRot(V:Boolean);
    procedure SetBoxColor1(V:TColor);
    procedure SetBoxColor2(V:TColor);
    function BlendColors(Back,Fore:TColor;A:Byte):TColor;
    function GetTabCount:Cardinal;
    procedure SetActiveTab(V:TMCoPXPTabSheet);
    function GetTab(Index:Cardinal):TMCoPXPTabSheet;
    function GetTabIndex:Integer;
    procedure SetTabIndex(V:Integer);
    procedure CalSheetRect;
    procedure SetTML(V:Integer);
    procedure SetTMR(V:Integer);
    procedure SetTMT(V:Integer);
    procedure SetTMB(V:Integer);
    procedure SetImages(V:TCustomImageList);
    procedure FlipBitmap(CW:Boolean;var B:{$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap);
    procedure SetImgT(V:Integer);
    procedure SetImgN(V:Integer);
    procedure SetScroll(V:Cardinal);
    function GetNextButtonVisible:Boolean;
    function GetPrevButtonVisible:Boolean;
    function GetScrollMax:Cardinal;
    procedure SetArrowColor(V:TColor);
    procedure SetOverColor(V:TColor);
    procedure MousePos(x,y:Integer);
    procedure SetCustomArrow(V:Boolean);
    procedure SetDrawArrow(V:TMCoPXPTabOnDrawArrow);
    procedure SetCloseColor(V:TColor);
    procedure SetClose(V:Boolean);
    procedure SetClsT(V:Integer);
    procedure SetClsN(V:Integer);
    procedure SetCustomClose(V:Boolean);
    procedure SetDrawClose(V:TMCoPXPTabOnDrawCloseButton);
    procedure SetCloseSize(V:TMCoPXPTabOnMeasureCloseButton);
    procedure SetArrowSize(V:TMCoPXPTabOnMeasureArrow);
    procedure SetColorScheme(V:TMCoPXPTabColorScheme);
    function StoreColor:Boolean;
    procedure CMMouseLeave(var Message:TMessage); message CM_MOUSELEAVE;
    function GetHomepage:String;
    procedure SetTabLine(V:TColor);
    procedure WMSetFocus(var Message:TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message:TWMSetFocus); message WM_KILLFOCUS;
    function TabVisible(Tab:TMCoPXPTabSheet):Boolean;
    procedure CNKeyDown(var Message:TWMKeyDown); message CN_KEYDOWN;
    procedure RefScroll;
    function CanChange:Boolean;
    procedure CMDialogChar(var Message:TCMDialogChar); message CM_DIALOGCHAR;
    procedure TCM_ADJUSTRECT(var Message:TMessage); message TCM_ADJUSTRECT;
    procedure TCM_DELETEALLITEMS(var Message:TMessage); message TCM_DELETEALLITEMS;
    procedure TCM_DELETEITEM(var Message:TMessage); message TCM_DELETEITEM;
    procedure TCM_GETCURFOCUS(var Message:TMessage); message TCM_GETCURFOCUS;
    procedure TCM_GETCURSEL(var Message:TMessage); message TCM_GETCURSEL;
    procedure TCM_GETIMAGELIST(var Message:TMessage); message TCM_GETIMAGELIST;
    procedure TCM_GETITEM(var Message:TMessage); message TCM_GETITEM ;
    procedure TCM_GETITEMCOUNT(var Message:TMessage); message TCM_GETITEMCOUNT;
    procedure TCM_GETITEMRECT(var Message:TMessage); message TCM_GETITEMRECT;
    procedure TCM_GETROWCOUNT(var Message:TMessage); message TCM_GETROWCOUNT;
    procedure TCM_GETTOOLTIPS(var Message:TMessage); message TCM_GETTOOLTIPS;
    procedure TCM_HITTEST(var Message:TMessage); message TCM_HITTEST;
    procedure TCM_INSERTITEM(var Message:TMessage); message TCM_INSERTITEM;
    procedure TCM_REMOVEIMAGE(var Message:TMessage); message TCM_REMOVEIMAGE;
    procedure TCM_SETCURFOCUS(var Message:TMessage); message TCM_SETCURFOCUS;
    procedure TCM_SETCURSEL(var Message:TMessage); message TCM_SETCURSEL;
    procedure TCM_SETIMAGELIST(var Message:TMessage); message TCM_SETIMAGELIST;
    procedure TCM_SETITEM(var Message:TMessage); message TCM_SETITEM;
    procedure TCM_SETITEMEXTRA(var Message:TMessage); message TCM_SETITEMEXTRA;
    procedure TCM_SETITEMSIZE(var Message:TMessage); message TCM_SETITEMSIZE;
    procedure TCM_SETPADDING(var Message:TMessage); message TCM_SETPADDING;
    procedure TCM_SETTOOLTIPS(var Message:TMessage); message TCM_SETTOOLTIPS;
    procedure TCM_SETMINTABWIDTH(var Message:TMessage); message TCM_SETMINTABWIDTH;
    procedure TCM_DESELECTALL(var Message:TMessage); message TCM_DESELECTALL;
    procedure TCM_HIGHLIGHTITEM(var Message:TMessage); message TCM_HIGHLIGHTITEM;
    procedure TCM_SETEXTENDEDSTYLE(var Message:TMessage); message TCM_SETEXTENDEDSTYLE;
    procedure TCM_GETEXTENDEDSTYLE(var Message:TMessage); message TCM_GETEXTENDEDSTYLE;
    procedure TCM_SETUNICODEFORMAT(var Message:TMessage); message TCM_SETUNICODEFORMAT;
    procedure TCM_GETUNICODEFORMAT(var Message:TMessage); message TCM_GETUNICODEFORMAT;
    procedure CMFontChanged(var Message:TMessage); message CM_FONTCHANGED;
    procedure ImageListChange(Sender:TObject);
    function GetVisibleTabsCount:Cardinal;
    procedure SetRTL(V:Boolean);
    function GetRect(Index:Cardinal):TRect;
    function GetArrowSize:Integer;
    function GetControl(Index:Integer):TControl;
    function GetControlCount:Integer;
    procedure ClsBtnClick(Sender:TObject);
    procedure ClsBtnEnter(Sender:TObject);
    procedure ClsBtnExit(Sender:TObject);
  protected
    SheetBack:{$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap;

    procedure Resize; override;
    procedure Paint; override;
    procedure Notification(AComponent:TComponent;Operation:TOperation); override;
    procedure MouseDown(Button:TMouseButton;Shift:TShiftState;X,Y:Integer); override;
    procedure MouseUp(Button:TMouseButton;Shift:TShiftState;X,Y:Integer);  override;
    procedure MouseMove(Shift:TShiftState;X,Y:Integer); override;
    procedure Ref;
    procedure RefTabs;
    procedure AdjustClientRect(var Rect:TRect); override;
    procedure RefActiveTab;
    procedure AssignTo(Dest:TPersistent); override;
    procedure DoCloseTab;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure RedrawControl;
    procedure EnterActivationCode(Name,Code:String);
    procedure Assign(Source:TPersistent); override;

    property TabCount:Cardinal read GetTabCount;
    property TabSheet[Index:Cardinal]:TMCoPXPTabSheet read GetTab;
    property SheetRect:TRect read fSheetRect;
    property ScrollPosition:Cardinal read fScroll write SetScroll;
    property NextButtonVisible:Boolean read GetNextButtonVisible;
    property PreviousButtonVisible:Boolean read GetPrevButtonVisible;
    property ScrollMax:Cardinal read GetScrollMax;
    property MouseOnTab:Integer read fOnTab;
    property VisibleTabsCount:Cardinal read GetVisibleTabsCount;
    property TabRect[Index:Cardinal]:TRect read GetRect;
    property ControlCount:Integer read GetControlCount;
    property Controls[Index:Integer]:TControl read GetControl;
  published
    property BackColor1:TColor read fBackColor1 write SetBackColor1 stored StoreColor;
    property BackColor2:TColor read fBackColor2 write SetBackColor2 stored StoreColor;
    property BackgroundGradientDirection:TMCoPXPTabGradientDirection read fBackDir write SetBackDir default gdHorizontal;
    property TabsWidth:Cardinal read fButtonWidth write SetButtonWidth default 150;
    property TabsHeight:Cardinal read fButtonHeight write SetButtonHeight default 40;
    property MarginLeft:Cardinal read fLMargin write SetLMargin default 8;
    property MarginRight:Cardinal read fRMargin write SetRMargin default 8;
    property MarginTop:Cardinal read fTMargin write SetTMargin default 8;
    property MarginBottom:Cardinal read fBMargin write SetBMargin default 8;
    property TabsPosition:TMCoPXPTabButtonsPosition read fButtonPos write SetButtonPos default bpAbove;
    property RotateTabs:Boolean read fButtonRot write SetButtonRot default False;
    property BoxColor1:TColor read fBoxColor1 write SetBoxColor1 stored StoreColor;
    property BoxColor2:TColor read fBoxColor2 write SetBoxColor2 stored StoreColor;
    property ActiveTab:TMCoPXPTabSheet read fActiveTab write SetActiveTab;
    property TabIndex:Integer read GetTabIndex write SetTabIndex stored False;
    property TextMarginLeft:Integer read fTML write SetTML default 40;
    property TextMarginRight:Integer read fTMR write SetTMR default 40;
    property TextMarginTop:Integer read fTMT write SetTMT default 15;
    property TextMarginBottom:Integer read fTMB write SetTMB default 5;
    property Images:TCustomImageList read fImages write SetImages;
    property ImageTangentialPosition:Integer read fImgT write SetImgT default 3;
    property ImageNormalPosition:Integer read fImgN write SetImgN default 4;
    property ArrowColor:TColor read fArrowColor write SetArrowColor stored StoreColor;
    property OverColor:TColor read fOverColor write SetOverColor stored StoreColor;
    property CustomArrow:Boolean read fCustomArrow write SetCustomArrow default False;
    property CloseButtonColor:TColor read fCloseColor write SetCloseColor stored StoreColor;
    property DrawCloseButton:Boolean read fClose write SetClose default False;
    property CloseButtonTangentialPosition:Integer read fClsT write SetClsT default 5;
    property CloseButtonNormalPosition:Integer read fClsN write SetClsN default 5;
    property CustomClose:Boolean read fCustomClose write SetCustomClose default False;
    property ColorScheme:TMCoPXPTabColorScheme read fColorScheme write SetColorScheme default csWindows;
    property Homepage:String read GetHomepage write fHomePage stored False;
    property FocusLineColor:TColor read fTabLine write SetTabLine stored StoreColor;
    property RTLReading:Boolean read fRTL write SetRTL default False;

    property Align;
    property Anchors;
    property BiDiMode;
    property Constraints;
    property Cursor;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property Height;
    property HelpContext;
    property HelpKeyword;
    property HelpType;
    property Hint;
    property Left;
    property Name;
    property ParentBiDiMode;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop default True;
    property Tag;
    property Top;
    property Visible;
    property Width;

    property OnDrawArrow:TMCoPXPTabOnDrawArrow read fDrawArrow write SetDrawArrow;
    property OnDrawCloseButton:TMCoPXPTabOnDrawCloseButton read fDrawClose write SetDrawClose;
    property OnMeasureCloseButton:TMCoPXPTabOnMeasureCloseButton read fCloseSize write SetCloseSize;
    property OnMeasureArrow:TMCoPXPTabOnMeasureArrow read fArrowSize write SetArrowSize;
    property OnCloseTab:TMCoPXPTabOnCloseTab read fCloseTab write fCloseTab;
    property OnChanging:TMCoPXPTabChanging read fChanging write fChanging;
    property OnChange:TNotifyEvent read fChange write fChange;
    property OnPaint:TNotifyEvent read fPaint write fPaint;

    property OnContextPopup;
    property OnDockDrop;
    property OnDockOver;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
  end;

const
  MCoPTabNone=-1;
  MCoPTabPreviousButton=-2;
  MCoPTabNextButton=-3;
  MCoPTabCloseButton=-4;

implementation

//Control

constructor TMCoPXPTabControl.Create(AOwner:TComponent);
begin
inherited Create(AOwner);
TabStop:=True;
Width:=289;
Height:=193;
ControlStyle:=[csDoubleClicks];
fBack:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
SheetBack:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
fBackDir:=gdHorizontal;
fButtonWidth:=150;
fButtonHeight:=40;
fLMargin:=8;
fRMargin:=8;
fTMargin:=8;
fBMargin:=8;
fButtonPos:=bpAbove;
fButtonRot:=False;
fActiveTab:=nil;
fTML:=40;
fTMR:=40;
fTMT:=15;
fTMB:=5;
fImages:=nil;
fImgT:=3;
fImgN:=4;
fScroll:=0;
fOnTab:=-1;
TabDown:=-1;
fCustomArrow:=False;
fDrawArrow:=nil;
MDown:=False;
fClsT:=5;
fClsN:=5;
fCustomClose:=False;
fDrawClose:=nil;
fCloseSize:=nil;
fArrowSize:=nil;
fCloseTab:=nil;
fChanging:=nil;
fChange:=nil;
fPaint:=nil;
fRTL:=False;
ClsBtn:=TButton.Create(nil);
ClsBtn.Top:=-10;
ClsBtn.Left:=-10;
ClsBtn.Width:=8;
ClsBtn.Height:=8;
ClsBtn.Parent:=Self;
ClsBtn.OnEnter:=ClsBtnEnter;
ClsBtn.OnExit:=ClsBtnExit;
ClsBtn.OnClick:=ClsBtnClick;
ClsBtn.Visible:=False;
ClsFoc:=False;
ImageChangeLink:=TChangeLink.Create;
ImageChangeLink.OnChange:=ImageListChange;
CalSheetRect;
SetColorScheme(csWindows);
end;

destructor TMCoPXPTabControl.Destroy;
begin
ClsBtn.Free;
SheetBack.Free;
fBack.Free;
ImageChangeLink.Free;
inherited Destroy;
end;

procedure TMCoPXPTabControl.Ref;
var
  b,z:{$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap;
  x,y,c,w,h,k,j,i,n,q,a:Integer;
  u:Cardinal;
  r:HRGN;
  s:String;
  l:TColor;
  e:TRect;
  d:TMCoPXPTabArrowDirection;
  t:TMCoPXPTabButtonState;

  procedure Arrow(ax,ay:Integer);
  begin
  fBack.Canvas.Pen.Color:=fArrowColor;
  case d of
    adUp:
      begin
      fBack.Canvas.MoveTo(ax,ay);
      fBack.Canvas.LineTo(ax-4,ay+4);
      fBack.Canvas.MoveTo(ax,ay);
      fBack.Canvas.LineTo(ax+4,ay+4);
      fBack.Canvas.MoveTo(ax,ay+1);
      fBack.Canvas.LineTo(ax-3,ay+4);
      fBack.Canvas.MoveTo(ax,ay+1);
      fBack.Canvas.LineTo(ax+3,ay+4);

      fBack.Canvas.MoveTo(ax,ay+4);
      fBack.Canvas.LineTo(ax-4,ay+4+4);
      fBack.Canvas.MoveTo(ax,ay+4);
      fBack.Canvas.LineTo(ax+4,ay+4+4);
      fBack.Canvas.MoveTo(ax,ay+4+1);
      fBack.Canvas.LineTo(ax-3,ay+4+4);
      fBack.Canvas.MoveTo(ax,ay+4+1);
      fBack.Canvas.LineTo(ax+3,ay+4+4);
      end;
    adDown:
      begin
      fBack.Canvas.MoveTo(ax,ay);
      fBack.Canvas.LineTo(ax-4,ay-4);
      fBack.Canvas.MoveTo(ax,ay);
      fBack.Canvas.LineTo(ax+4,ay-4);
      fBack.Canvas.MoveTo(ax,ay-1);
      fBack.Canvas.LineTo(ax-3,ay-4);
      fBack.Canvas.MoveTo(ax,ay-1);
      fBack.Canvas.LineTo(ax+3,ay-4);

      fBack.Canvas.MoveTo(ax,ay-4);
      fBack.Canvas.LineTo(ax-4,ay-4-4);
      fBack.Canvas.MoveTo(ax,ay-4);
      fBack.Canvas.LineTo(ax+4,ay-4-4);
      fBack.Canvas.MoveTo(ax,ay-4-1);
      fBack.Canvas.LineTo(ax-3,ay-4-4);
      fBack.Canvas.MoveTo(ax,ay-4-1);
      fBack.Canvas.LineTo(ax+3,ay-4-4);
      end;
    adLeft:
      begin
      fBack.Canvas.MoveTo(ax,ay);
      fBack.Canvas.LineTo(ax+4,ay-4);
      fBack.Canvas.MoveTo(ax,ay);
      fBack.Canvas.LineTo(ax+4,ay+4);
      fBack.Canvas.MoveTo(ax+1,ay);
      fBack.Canvas.LineTo(ax+4,ay-3);
      fBack.Canvas.MoveTo(ax+1,ay);
      fBack.Canvas.LineTo(ax+4,ay+3);

      fBack.Canvas.MoveTo(ax+4,ay);
      fBack.Canvas.LineTo(ax+4+4,ay-4);
      fBack.Canvas.MoveTo(ax+4,ay);
      fBack.Canvas.LineTo(ax+4+4,ay+4);
      fBack.Canvas.MoveTo(ax+1+4,ay);
      fBack.Canvas.LineTo(ax+4+4,ay-3);
      fBack.Canvas.MoveTo(ax+1+4,ay);
      fBack.Canvas.LineTo(ax+4+4,ay+3);
      end;
    adRight:
      begin
      fBack.Canvas.MoveTo(ax,ay);
      fBack.Canvas.LineTo(ax-4,ay-4);
      fBack.Canvas.MoveTo(ax,ay);
      fBack.Canvas.LineTo(ax-4,ay+4);
      fBack.Canvas.MoveTo(ax-1,ay);
      fBack.Canvas.LineTo(ax-4,ay-3);
      fBack.Canvas.MoveTo(ax-1,ay);
      fBack.Canvas.LineTo(ax-4,ay+3);

      fBack.Canvas.MoveTo(ax-4,ay);
      fBack.Canvas.LineTo(ax-4-4,ay-4);
      fBack.Canvas.MoveTo(ax-4,ay);
      fBack.Canvas.LineTo(ax-4-4,ay+4);
      fBack.Canvas.MoveTo(ax-1-4,ay);
      fBack.Canvas.LineTo(ax-4-4,ay-3);
      fBack.Canvas.MoveTo(ax-1-4,ay);
      fBack.Canvas.LineTo(ax-4-4,ay+3);
      end;
  end;
  end;

  procedure CloseBtn(ax,ay:Integer);
  begin
  fBack.Canvas.Pen.Color:=fCloseColor;
  fBack.Canvas.MoveTo(ax,ay);
  fBack.Canvas.LineTo(ax+7,ay+7);
  fBack.Canvas.MoveTo(ax+1,ay);
  fBack.Canvas.LineTo(ax+8,ay+7);
  fBack.Canvas.MoveTo(ax+6,ay);
  fBack.Canvas.LineTo(ax-1,ay+7);
  fBack.Canvas.MoveTo(ax+7,ay);
  fBack.Canvas.LineTo(ax,ay+7);
  if not ClsFoc then Exit;
  fBack.Canvas.Pen.Style:=psDot;
  fBack.Canvas.Brush.Style:=bsClear;
  fBack.Canvas.Rectangle(ax-2,ay-2,ax+10,ay+9);
  fBack.Canvas.Pen.Style:=psSolid;
  fBack.Canvas.Brush.Style:=bsSolid;
  end;

begin
if csDestroying in ComponentState then Exit;
if fBack=nil then Exit;
fBack.Width:=Width;
fBack.Height:=Height;
Gradient(fBackColor1,fBackColor2,fBackDir=gdHorizontal,fBack);

b:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
b.Width:=Width-Integer(fLMargin+fRMargin);
b.Height:=Height-Integer(fTMargin+fBMargin);
x:=fLMargin;
y:=fTMargin;
case fButtonPos of
  bpLeft:
    if fButtonRot then
      begin
      Inc(x,fButtonHeight);
      b.Width:=b.Width-Integer(fButtonHeight);
      end
    else
      begin
      Inc(x,fButtonWidth);
      b.Width:=b.Width-Integer(fButtonWidth);
      end;
  bpRight:
    if fButtonRot then
      b.Width:=b.Width-Integer(fButtonHeight)
    else
      b.Width:=b.Width-Integer(fButtonWidth);
  bpAbove:
    begin
    Inc(y,fButtonHeight);
    b.Height:=b.Height-Integer(fButtonHeight);
    end;
  bpBelow:b.Height:=b.Height-Integer(fButtonHeight);
end;
Gradient(fBoxColor1,fBoxColor2,(fButtonPos=bpAbove) or (fButtonPos=bpBelow),b);
r:=CreateRoundRectRgn(x,y,x+b.Width,y+b.Height,5,5);
SelectClipRGN(fBack.Canvas.Handle,r);
fBack.Canvas.Draw(x,y,b);
DeleteObject(r);
SelectClipRGN(fBack.Canvas.Handle,0);
fBack.Canvas.Pixels[x+1,y]:=BlendColors(fBack.Canvas.Pixels[x+1,y],b.Canvas.Pixels[1,0],128);
fBack.Canvas.Pixels[x,y+1]:=BlendColors(fBack.Canvas.Pixels[x,y+1],b.Canvas.Pixels[0,1],128);
fBack.Canvas.Pixels[x+b.Width-3,y]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-3,y],b.Canvas.Pixels[b.Width-2,0],128);
fBack.Canvas.Pixels[x+b.Width-2,y+1]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-2,y+1],b.Canvas.Pixels[b.Width-1,1],128);
fBack.Canvas.Pixels[x+1,y+b.Height-2]:=BlendColors(fBack.Canvas.Pixels[x+1,y+b.Height-2],b.Canvas.Pixels[1,b.Height-1],128);
fBack.Canvas.Pixels[x,y+b.Height-3]:=BlendColors(fBack.Canvas.Pixels[x,y+b.Height-3],b.Canvas.Pixels[0,b.Height-2],128);
fBack.Canvas.Pixels[x+b.Width-3,y+b.Height-2]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-3,y+b.Height-2],b.Canvas.Pixels[b.Width-2,b.Height-1],128);
fBack.Canvas.Pixels[x+b.Width-2,y+b.Height-3]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-2,y+b.Height-3],b.Canvas.Pixels[b.Width-1,b.Height-2],128);
if fBackDir=gdHorizontal then
  begin
  fBack.Canvas.Pixels[x+2,y+b.Height-1]:=BlendColors(fBack.Canvas.Pixels[x,y+b.Height-1],clBlack,50);
  fBack.Canvas.Pen.Color:=BlendColors(fBack.Canvas.Pixels[x,y+b.Height-1],clBlack,75);
  fBack.Canvas.MoveTo(x+3,y+b.Height-1);
  fBack.Canvas.LineTo(x+b.Width-1,y+b.Height-1);

  fBack.Canvas.Pixels[x+3,y+b.Height]:=BlendColors(fBack.Canvas.Pixels[x,y+b.Height],clBlack,25);
  fBack.Canvas.Pen.Color:=BlendColors(fBack.Canvas.Pixels[x,y+b.Height],clBlack,50);
  fBack.Canvas.MoveTo(x+4,y+b.Height);
  fBack.Canvas.LineTo(x+b.Width-1,y+b.Height);

  fBack.Canvas.Pixels[x+b.Width-1,y+2]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-1,y+2],clBlack,50);
  fBack.Canvas.Pixels[x+b.Width,y+3]:=BlendColors(fBack.Canvas.Pixels[x+b.Width,y+3],clBlack,25);
  for c:=3 to b.Height-2 do
    begin
    fBack.Canvas.Pixels[x+b.Width-1,y+c]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-1,y+c],clBlack,75);
    if c<>3 then fBack.Canvas.Pixels[x+b.Width,y+c]:=BlendColors(fBack.Canvas.Pixels[x+b.Width,y+c],clBlack,50);
    end;
  fBack.Canvas.Pixels[x+b.Width-2,y+b.Height-2]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-2,y+b.Height-2],clBlack,75);
  fBack.Canvas.Pixels[x+b.Width-1,y+b.Height-1]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-1,y+b.Height-1],clBlack,50);
  fBack.Canvas.Pixels[x+b.Width,y+b.Height-1]:=BlendColors(fBack.Canvas.Pixels[x+b.Width,y+b.Height-1],clBlack,25);
  fBack.Canvas.Pixels[x+b.Width-1,y+b.Height]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-1,y+b.Height],clBlack,25);
  end
else
  begin
  fBack.Canvas.Pixels[x+b.Width-1,y+2]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-1,y],clBlack,50);
  fBack.Canvas.Pen.Color:=BlendColors(fBack.Canvas.Pixels[x+b.Width-1,y],clBlack,75);
  fBack.Canvas.MoveTo(x+b.Width-1,y+3);
  fBack.Canvas.LineTo(x+b.Width-1,y+b.Height-1);

  fBack.Canvas.Pixels[x+b.Width,y+3]:=BlendColors(fBack.Canvas.Pixels[x+b.Width,y],clBlack,25);
  fBack.Canvas.Pen.Color:=BlendColors(fBack.Canvas.Pixels[x+b.Width,y],clBlack,50);
  fBack.Canvas.MoveTo(x+b.Width,y+4);
  fBack.Canvas.LineTo(x+b.Width,y+b.Height-1);

  fBack.Canvas.Pixels[x+2,y+b.Height-1]:=BlendColors(fBack.Canvas.Pixels[x+2,y+b.Height-1],clBlack,50);
  fBack.Canvas.Pixels[x+3,y+b.Height]:=BlendColors(fBack.Canvas.Pixels[x+3,y+b.Height],clBlack,25);
  for c:=3 to b.Width-2 do
    begin
    fBack.Canvas.Pixels[x+c,y+b.Height-1]:=BlendColors(fBack.Canvas.Pixels[x+c,y+b.Height-1],clBlack,75);
    if c<>3 then fBack.Canvas.Pixels[x+c,y+b.Height]:=BlendColors(fBack.Canvas.Pixels[x+c,y+b.Height],clBlack,50);
    end;
  fBack.Canvas.Pixels[x+b.Width-2,y+b.Height-2]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-2,y+b.Height-2],clBlack,75);
  fBack.Canvas.Pixels[x+b.Width-1,y+b.Height-1]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-1,y+b.Height-1],clBlack,50);
  fBack.Canvas.Pixels[x+b.Width,y+b.Height-1]:=BlendColors(fBack.Canvas.Pixels[x+b.Width,y+b.Height-1],clBlack,25);
  fBack.Canvas.Pixels[x+b.Width-1,y+b.Height]:=BlendColors(fBack.Canvas.Pixels[x+b.Width-1,y+b.Height],clBlack,25);
  end;
b.Free;

case fButtonPos of
  bpLeft:x:=fLMargin;
  bpRight:x:=Width-Integer(fRMargin);
  bpAbove,bpBelow:
    if BiDiMode=bdRightToLeft then
      x:=Width-Integer(fRMargin)-4
    else
      x:=fLMargin+4;
end;
case fButtonPos of
  bpAbove:y:=fTMargin;
  bpBelow:y:=Height-Integer(fBMargin);
  bpLeft,bpRight:y:=fTMargin+4;
end;
if ((fButtonPos=bpLeft) or (fButtonPos=bpRight)) and (fButtonRot) then
  begin
  w:=fButtonHeight;
  h:=fButtonWidth;
  end
else
  begin
  w:=fButtonWidth;
  h:=fButtonHeight;
  end;
if ((BiDiMode=bdRightToLeft) and (fButtonPos<>bpLeft)) or (fButtonPos=bpRight) then
  begin
  w:=-w;
  Dec(x,1);
  end;
if fButtonPos=bpBelow then
  begin
  h:=-h;
  Dec(y,1);
  end;
if GetPrevButtonVisible then
  begin
  if BiDiMode=bdRightToLeft then i:=-1 else i:=1;
  case fButtonPos of
    bpAbove,bpBelow:Inc(x,GetArrowSize*i);
    bpLeft,bpRight:Inc(y,GetArrowSize);
  end;
  end;
if GetNextButtonVisible then i:=GetArrowSize else i:=0;
z:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
if BiDiMode=bdRightToLeft then
  begin
  case fButtonPos of
    bpAbove:
      begin
      z.Width:=Integer(fLMargin)+5+i;
      z.Height:=fTMargin+fButtonHeight;
      end;
    bpBelow:
      begin
      z.Width:=Integer(fLMargin)+5+i;
      z.Height:=fBMargin+fButtonHeight;
      end;
    bpLeft:
      begin
      if fButtonRot then
        z.Width:=fLMargin+fButtonHeight
      else
        z.Width:=fLMargin+fButtonWidth;
      z.Height:=Integer(fBMargin)+5+i;
      end;
    bpRight:
      begin
      if fButtonRot then
        z.Width:=fRMargin+fButtonHeight
      else
        z.Width:=fRMargin+fButtonWidth;
      z.Height:=Integer(fBMargin)+5+i;
      end;
  end;
  end
else
  begin
  case fButtonPos of
    bpAbove:
      begin
      z.Width:=Integer(fRMargin)+5+i;
      z.Height:=fTMargin+fButtonHeight;
      end;
    bpBelow:
      begin
      z.Width:=Integer(fRMargin)+5+i;
      z.Height:=fBMargin+fButtonHeight;
      end;
    bpLeft:
      begin
      if fButtonRot then
        z.Width:=fLMargin+fButtonHeight
      else
        z.Width:=fLMargin+fButtonWidth;
      z.Height:=Integer(fBMargin)+5+i;
      end;
    bpRight:
      begin
      if fButtonRot then
        z.Width:=fRMargin+fButtonHeight
      else
        z.Width:=fRMargin+fButtonWidth;
      z.Height:=Integer(fBMargin)+5+i;
      end;
  end;
  end;
case fButtonPos of
  bpLeft:z.Canvas.Draw(0,z.Height-Height,fBack);
  bpRight:z.Canvas.Draw(z.Width-Width+1,z.Height-Height,fBack);
  bpAbove:
    if BiDiMode=bdRightToLeft then
      z.Canvas.Draw(0,0,fBack)
    else
      z.Canvas.Draw(z.Width-Width,0,fBack);
  bpBelow:
    if BiDiMode=bdRightToLeft then
      z.Canvas.Draw(0,z.Height-Height+1,fBack)
    else
      z.Canvas.Draw(z.Width-Width,z.Height-Height+1,fBack);
end;
u:=0;
if fScroll>0 then
  begin
  u:=fScroll;
  for c:=0 to ControlCount-1 do
    begin
    if TabVisible(TMCoPXPTabSheet(Controls[c])) then dec(u);
    if u=0 then
      begin
      u:=c+1;
      Break;
      end;
    end;
  end;
for c:=u to ControlCount-1 do
  begin
  if not TabVisible(TMCoPXPTabSheet(Controls[c])) then Continue;
  if Controls[c]=fActiveTab then
    begin
    case fButtonPos of
      bpAbove,bpLeft:fBack.Canvas.Brush.Color:=fBoxColor1;
      bpBelow,bpRight:fBack.Canvas.Brush.Color:=fBoxColor2;
    end;
    fBack.Canvas.FillRect(Rect(x,y,x+w,y+h));
    if fButtonPos in [bpAbove,bpLeft] then
      begin
      if w>0 then k:=x else k:=x+w;
      if h>0 then j:=y else j:=y+h;
      if fBackDir=gdHorizontal then
        begin
        i:=-1;
        n:=0;
        end
      else
        begin
        i:=0;
        n:=-1;
        end;
      fBack.Canvas.Pixels[k,j]:=fBack.Canvas.Pixels[k+i,j+n];
      fBack.Canvas.Pixels[k,j+1]:=BlendColors(fBack.Canvas.Pixels[k+i,j+n+1],fBack.Canvas.Brush.Color,128);
      fBack.Canvas.Pixels[k+1,j]:=BlendColors(fBack.Canvas.Pixels[k+i+1,j+n],fBack.Canvas.Brush.Color,128);
      end;
    if fButtonPos in [bpAbove,bpRight] then
      begin
      if w>0 then k:=x+w-1 else k:=x-1;
      if h>0 then j:=y else j:=y+h;
      if fBackDir=gdHorizontal then
        begin
        i:=1;
        n:=0;
        end
      else
        begin
        i:=0;
        n:=-1;
        end;
      fBack.Canvas.Pixels[k,j]:=fBack.Canvas.Pixels[k+i,j+n];
      fBack.Canvas.Pixels[k,j+1]:=BlendColors(fBack.Canvas.Pixels[k+i,j+n+1],fBack.Canvas.Brush.Color,128);
      fBack.Canvas.Pixels[k-1,j]:=BlendColors(fBack.Canvas.Pixels[k+i-1,j+n],fBack.Canvas.Brush.Color,128);
      end;
    if fButtonPos in [bpBelow,bpLeft] then
      begin
      if w>0 then k:=x else k:=x+w;
      if h>0 then j:=y+h-1 else j:=y-1;
      if fBackDir=gdHorizontal then
        begin
        i:=-1;
        n:=0;
        end
      else
        begin
        i:=0;
        n:=1;
        end;
      fBack.Canvas.Pixels[k,j]:=fBack.Canvas.Pixels[k+i,j+n];
      fBack.Canvas.Pixels[k,j-1]:=BlendColors(fBack.Canvas.Pixels[k+i,j+n-1],fBack.Canvas.Brush.Color,128);
      fBack.Canvas.Pixels[k+1,j]:=BlendColors(fBack.Canvas.Pixels[k+i+1,j+n],fBack.Canvas.Brush.Color,128);
      end;
    if fButtonPos in [bpBelow,bpRight] then
      begin
      if w>0 then k:=x+w-1 else k:=x-1;
      if h>0 then j:=y+h-1 else j:=y-1;
      if fBackDir=gdHorizontal then
        begin
        i:=1;
        n:=0;
        end
      else
        begin
        i:=0;
        n:=1;
        end;
      fBack.Canvas.Pixels[k,j]:=fBack.Canvas.Pixels[k+i,j+n];
      fBack.Canvas.Pixels[k,j-1]:=BlendColors(fBack.Canvas.Pixels[k+i,j+n-1],fBack.Canvas.Brush.Color,128);
      fBack.Canvas.Pixels[k-1,j]:=BlendColors(fBack.Canvas.Pixels[k+i-1,j+n],fBack.Canvas.Brush.Color,128);
      end;
    case fButtonPos of
      bpAbove:
        begin
        fBack.Canvas.Pixels[x-1,y+h-1]:=BlendColors(fBack.Canvas.Pixels[x-1,y+h-1],fBack.Canvas.Brush.Color,128);
        fBack.Canvas.Pixels[x+w,y+h-1]:=BlendColors(fBack.Canvas.Pixels[x+w,y+h-1],fBack.Canvas.Brush.Color,128);
        end;
      bpBelow:
        begin
        fBack.Canvas.Pixels[x-1,y+h]:=BlendColors(fBack.Canvas.Pixels[x-1,y+h],fBack.Canvas.Brush.Color,128);
        fBack.Canvas.Pixels[x+w,y+h]:=BlendColors(fBack.Canvas.Pixels[x+w,y+h],fBack.Canvas.Brush.Color,128);
        end;
      bpLeft:
        begin
        fBack.Canvas.Pixels[x+w-1,y-1]:=BlendColors(fBack.Canvas.Pixels[x+w-1,y-1],fBack.Canvas.Brush.Color,128);
        fBack.Canvas.Pixels[x+w-1,y+h]:=BlendColors(fBack.Canvas.Pixels[x+w-1,y+h],fBack.Canvas.Brush.Color,128);
        end;
      bpRight:
        begin
        fBack.Canvas.Pixels[x+w,y-1]:=BlendColors(fBack.Canvas.Pixels[x+w,y-1],fBack.Canvas.Brush.Color,128);
        fBack.Canvas.Pixels[x+w,y+h]:=BlendColors(fBack.Canvas.Pixels[x+w,y+h],fBack.Canvas.Brush.Color,128);
        end;
    end;
    if fButtonPos<>bpLeft then
      begin
      if w>0 then k:=x+w else k:=x;
      if h>0 then j:=y+2 else j:=y+h+2;
      if fButtonPos=bpBelow then
        Dec(j,2)
      else
        fBack.Canvas.Pixels[k,j]:=BlendColors(fBack.Canvas.Pixels[k,j],clBlack,50);
      fBack.Canvas.Pixels[k+1,j+1]:=BlendColors(fBack.Canvas.Pixels[k+1,j+1],clBlack,25);
      n:=j+Abs(h)-3;
      if fButtonPos=bpAbove then Dec(n);
      for i:=j+1 to n do
        fBack.Canvas.Pixels[k,i]:=BlendColors(fBack.Canvas.Pixels[k,i],clBlack,75);
      for i:=j+2 to j+Abs(h)-3 do
        fBack.Canvas.Pixels[k+1,i]:=BlendColors(fBack.Canvas.Pixels[k+1,i],clBlack,50);
      if fButtonPos=bpBelow then
        begin
        fBack.Canvas.Pixels[k,j+Abs(h)-2]:=BlendColors(fBack.Canvas.Pixels[k,j+Abs(h)-2],clBlack,75);
        fBack.Canvas.Pixels[k,j+Abs(h)-1]:=BlendColors(fBack.Canvas.Pixels[k,j+Abs(h)-1],clBlack,75);
        fBack.Canvas.Pixels[k+1,j+Abs(h)-2]:=BlendColors(fBack.Canvas.Pixels[k+1,j+Abs(h)-2],clBlack,50);
        fBack.Canvas.Pixels[k+1,j+Abs(h)-1]:=BlendColors(fBack.Canvas.Pixels[k+1,j+Abs(h)-1],clBlack,50);
        end;
      end;
    if fButtonPos<>bpAbove then
      begin
      if w>0 then k:=x+2 else k:=x+w+2;
      if h>0 then j:=y+h else j:=y;
      if fButtonPos=bpRight then
        Dec(k,2)
      else
        fBack.Canvas.Pixels[k,j]:=BlendColors(fBack.Canvas.Pixels[k,j],clBlack,50);
      fBack.Canvas.Pixels[k+1,j+1]:=BlendColors(fBack.Canvas.Pixels[k+1,j+1],clBlack,25);
      n:=k+Abs(w)-3;
      if fButtonPos=bpLeft then Dec(n);
      for i:=k+1 to n do
        fBack.Canvas.Pixels[i,j]:=BlendColors(fBack.Canvas.Pixels[i,j],clBlack,75);
      for i:=k+2 to k+Abs(w)-3 do
        fBack.Canvas.Pixels[i,j+1]:=BlendColors(fBack.Canvas.Pixels[i,j+1],clBlack,50);
      if fButtonPos=bpRight then
        begin
        fBack.Canvas.Pixels[k+Abs(w)-2,j]:=BlendColors(fBack.Canvas.Pixels[k+Abs(w)-2,j],clBlack,75);
        fBack.Canvas.Pixels[k+Abs(w)-1,j]:=BlendColors(fBack.Canvas.Pixels[k+Abs(w)-1,j],clBlack,75);
        fBack.Canvas.Pixels[k+Abs(w)-2,j+1]:=BlendColors(fBack.Canvas.Pixels[k+Abs(w)-2,j+1],clBlack,50);
        fBack.Canvas.Pixels[k+Abs(w)-1,j+1]:=BlendColors(fBack.Canvas.Pixels[k+Abs(w)-1,j+1],clBlack,50);
        end;
      end;
    if (fButtonPos=bpBelow) or (fButtonPos=bpRight) then
      begin
      if w>0 then k:=x+w else k:=x;
      if h>0 then j:=y+h else j:=y;
      fBack.Canvas.Pixels[k-1,j-1]:=BlendColors(fBack.Canvas.Pixels[k-1,j-1],clBlack,75);
      fBack.Canvas.Pixels[k,j]:=BlendColors(fBack.Canvas.Pixels[k,j],clBlack,50);
      fBack.Canvas.Pixels[k,j+1]:=BlendColors(fBack.Canvas.Pixels[k,j+1],clBlack,25);
      fBack.Canvas.Pixels[k+1,j]:=BlendColors(fBack.Canvas.Pixels[k+1,j],clBlack,25);
      end;
    if fClose then
      begin
      i:=8;
      n:=8;
      if fCustomClose then
        if Assigned(fCloseSize) then
          fCloseSize(Self,i,n);
      i:=Abs(i);
      n:=Abs(n);
      if w>0 then k:=x+w-fClsT else k:=x-fClsT;
      if ((fButtonPos in [bpAbove,bpBelow]) or (not fButtonRot)) and (BiDiMode=bdRightToLeft) then
        k:=k-Abs(w)+2*fClsT+i;
      if (fButtonRot) and (fButtonPos=bpLeft) then
        k:=k-Abs(w)+2*fClsT+i;
      if h>0 then j:=y+fClsN else j:=y+h+fClsN;
      if fButtonRot then
      if ((fButtonPos=bpLeft) and (BiDiMode=bdRightToLeft)) or
         ((fButtonPos=bpRight) and (BiDiMode<>bdRightToLeft)) then
        j:=j+Abs(h)-2*fClsN-n;
      if fCustomClose then
        begin
        t:=bsNormal;
        if fOnTab=MCoPTabCloseButton then t:=bsOver;
        if (MDown) and (fOnTab=MCoPTabCloseButton) then t:=bsPressed;
        if Assigned(fDrawClose) then
          fDrawClose(Self,fBack.Canvas,k-i,j,t,ClsFoc);
        end
      else
        CloseBtn(k-i,j);
      CloseButtonRect:=Bounds(k-i,j,i,n);
      end;
    if Focused then
      begin
      if w>0 then k:=x else k:=x+w;
      if h>0 then j:=y else j:=y+h;
      fBack.Canvas.Pen.Color:=fTabLine;
      case fButtonPos of
        bpAbove:
          begin
          fBack.Canvas.MoveTo(k+2,j+1);
          fBack.Canvas.LineTo(k+Abs(w)-2,j+1);
          end;
        bpBelow:
          begin
          fBack.Canvas.MoveTo(k+2,j+Abs(h)-2);
          fBack.Canvas.LineTo(k+Abs(w)-2,j+Abs(h)-2);
          end;
        bpLeft:
          begin
          fBack.Canvas.MoveTo(k+1,j+2);
          fBack.Canvas.LineTo(k+1,j+Abs(h)-2);
          end;
        bpRight:
          begin
          fBack.Canvas.MoveTo(k+Abs(w)-2,j+2);
          fBack.Canvas.LineTo(k+Abs(w)-2,j+Abs(h)-2);
          end;
      end;
      end;
    end
  else if c=fOnTab then
    begin
    fBack.Canvas.Brush.Color:=fOverColor;
    fBack.Canvas.FillRect(Rect(x,y,x+w,y+h));
    if fButtonPos in [bpAbove,bpLeft] then
      begin
      if w>0 then k:=x else k:=x+w;
      if h>0 then j:=y else j:=y+h;
      if fBackDir=gdHorizontal then
        begin
        i:=-1;
        n:=0;
        end
      else
        begin
        i:=0;
        n:=-1;
        end;
      fBack.Canvas.Pixels[k,j]:=fBack.Canvas.Pixels[k+i,j+n];
      fBack.Canvas.Pixels[k,j+1]:=BlendColors(fBack.Canvas.Pixels[k+i,j+n+1],fBack.Canvas.Brush.Color,128);
      fBack.Canvas.Pixels[k+1,j]:=BlendColors(fBack.Canvas.Pixels[k+i+1,j+n],fBack.Canvas.Brush.Color,128);
      end;
    if fButtonPos in [bpAbove,bpRight] then
      begin
      if w>0 then k:=x+w-1 else k:=x-1;
      if h>0 then j:=y else j:=y+h;
      if fBackDir=gdHorizontal then
        begin
        i:=1;
        n:=0;
        end
      else
        begin
        i:=0;
        n:=-1;
        end;
      fBack.Canvas.Pixels[k,j]:=fBack.Canvas.Pixels[k+i,j+n];
      fBack.Canvas.Pixels[k,j+1]:=BlendColors(fBack.Canvas.Pixels[k+i,j+n+1],fBack.Canvas.Brush.Color,128);
      fBack.Canvas.Pixels[k-1,j]:=BlendColors(fBack.Canvas.Pixels[k+i-1,j+n],fBack.Canvas.Brush.Color,128);
      end;
    if fButtonPos in [bpBelow,bpLeft] then
      begin
      if w>0 then k:=x else k:=x+w;
      if h>0 then j:=y+h-1 else j:=y-1;
      if fBackDir=gdHorizontal then
        begin
        i:=-1;
        n:=0;
        end
      else
        begin
        i:=0;
        n:=1;
        end;
      fBack.Canvas.Pixels[k,j]:=fBack.Canvas.Pixels[k+i,j+n];
      fBack.Canvas.Pixels[k,j-1]:=BlendColors(fBack.Canvas.Pixels[k+i,j+n-1],fBack.Canvas.Brush.Color,128);
      fBack.Canvas.Pixels[k+1,j]:=BlendColors(fBack.Canvas.Pixels[k+i+1,j+n],fBack.Canvas.Brush.Color,128);
      end;
    if fButtonPos in [bpBelow,bpRight] then
      begin
      if w>0 then k:=x+w-1 else k:=x-1;
      if h>0 then j:=y+h-1 else j:=y-1;
      if fBackDir=gdHorizontal then
        begin
        i:=1;
        n:=0;
        end
      else
        begin
        i:=0;
        n:=1;
        end;
      fBack.Canvas.Pixels[k,j]:=fBack.Canvas.Pixels[k+i,j+n];
      fBack.Canvas.Pixels[k,j-1]:=BlendColors(fBack.Canvas.Pixels[k+i,j+n-1],fBack.Canvas.Brush.Color,128);
      fBack.Canvas.Pixels[k-1,j]:=BlendColors(fBack.Canvas.Pixels[k+i-1,j+n],fBack.Canvas.Brush.Color,128);
      end;
    case fButtonPos of
      bpAbove:
        begin
        fBack.Canvas.Pixels[x-1,y+h-1]:=BlendColors(fBack.Canvas.Pixels[x-1,y+h-1],fBack.Canvas.Brush.Color,128);
        fBack.Canvas.Pixels[x+w,y+h-1]:=BlendColors(fBack.Canvas.Pixels[x+w,y+h-1],fBack.Canvas.Brush.Color,128);
        end;
      bpBelow:
        begin
        fBack.Canvas.Pixels[x-1,y+h]:=BlendColors(fBack.Canvas.Pixels[x-1,y+h],fBack.Canvas.Brush.Color,128);
        fBack.Canvas.Pixels[x+w,y+h]:=BlendColors(fBack.Canvas.Pixels[x+w,y+h],fBack.Canvas.Brush.Color,128);
        end;
      bpLeft:
        begin
        fBack.Canvas.Pixels[x+w-1,y-1]:=BlendColors(fBack.Canvas.Pixels[x+w-1,y-1],fBack.Canvas.Brush.Color,128);
        fBack.Canvas.Pixels[x+w-1,y+h]:=BlendColors(fBack.Canvas.Pixels[x+w-1,y+h],fBack.Canvas.Brush.Color,128);
        end;
      bpRight:
        begin
        fBack.Canvas.Pixels[x+w,y-1]:=BlendColors(fBack.Canvas.Pixels[x+w,y-1],fBack.Canvas.Brush.Color,128);
        fBack.Canvas.Pixels[x+w,y+h]:=BlendColors(fBack.Canvas.Pixels[x+w,y+h],fBack.Canvas.Brush.Color,128);
        end;
    end;
    end;
  s:=TMCoPXPTabSheet(Controls[c]).Caption;
  b:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
  if (fButtonRot) and (fButtonPos in [bpLeft,bpRight]) then
    begin
    b.Width:=Abs(w)-fTMT-fTMB;
    b.Height:=Abs(h)-fTML-fTMR;
    k:=0;
    j:=0;
    case fButtonPos of
      bpLeft:
        begin
        if w>0 then k:=x+fTMT else k:=x+w+fTMT;
        if h>0 then j:=y+fTMR else j:=y+h+fTMR;
        end;
      bpRight:
        begin
        if w>0 then k:=x+fTMB else k:=x+w+fTMB;
        if h>0 then j:=y+fTML else j:=y+h+fTML;
        end;
    end;
    end
  else
    begin
    b.Width:=Abs(w)-fTML-fTMR;
    b.Height:=Abs(h)-fTMT-fTMB;
    if w>0 then k:=x+fTML else k:=x+w+fTML;
    if h>0 then j:=y+fTMT else j:=y+h+fTMT;
    end;
  b.Canvas.CopyRect(b.Canvas.ClipRect,fBack.Canvas,Bounds(k,j,b.Width,b.Height));
  if (fButtonRot) and (fButtonPos=bpLeft) then FlipBitmap(True,b);
  if (fButtonRot) and (fButtonPos=bpRight) then FlipBitmap(False,b);
  b.Canvas.Brush.Style:=bsClear;
  b.Canvas.Font.Assign(Font);
  e:=b.Canvas.ClipRect;
  if BiDiMode=bdLeftToRight then
    i:=DT_LEFT
  else
    i:=DT_RIGHT;
  if fRTL then i:=i or DT_RTLREADING;
  DrawText(b.Canvas.Handle,PChar(s),Length(s),e,i);
  if (fButtonRot) and (fButtonPos=bpLeft) then FlipBitmap(False,b);
  if (fButtonRot) and (fButtonPos=bpRight) then FlipBitmap(True,b);
  fBack.Canvas.Draw(k,j,b);
  b.Free;
  if Assigned(fImages) and (TMCoPXPTabSheet(Controls[c]).ImageIndex<>-1) then
    begin
    if (fButtonRot) and (fButtonPos in [bpLeft,bpRight]) then
      begin
      if w>0 then k:=x+fImgN else k:=x-fImgN-fImages.Width;
      if h>0 then j:=y+fImgT else j:=y+h+fImgT;
      if ((fButtonPos=bpLeft) and (BiDiMode<>bdRightToLeft)) or
         ((fButtonPos=bpRight) and (BiDiMode=bdRightToLeft)) then
        j:=y+h-fImgT-fImages.Height;
      end
    else
      begin
      if w>0 then k:=x+fImgT else k:=x+w+fImgT;
      if h>0 then j:=y+fImgN else j:=y+h+fImgN;
      if (BiDiMode=bdRightToLeft) and (w>0) then k:=x+w-fImgT-fImages.Width;
      if (BiDiMode=bdRightToLeft) and (w<0) then k:=x-fImgT-fImages.Width;
      end;
    if TMCoPXPTabSheet(Controls[c]).ImageIndex<fImages.Count then
      fImages.Draw(fBack.Canvas,k,j,TMCoPXPTabSheet(Controls[c]).ImageIndex);
    end;
  case fButtonPos of
    bpAbove,bpBelow:Inc(x,w+(Abs(w) div w));
    bpLeft,bpRight:Inc(y,h+(Abs(h) div h));
  end;
  end;
case fButtonPos of
  bpLeft:fBack.Canvas.Draw(0,Height-z.Height,z);
  bpRight:fBack.Canvas.Draw(Width-z.Width-1,Height-z.Height,z);
  bpAbove:
    if BiDiMode=bdRightToLeft then
      fBack.Canvas.Draw(0,0,z)
    else
      fBack.Canvas.Draw(Width-z.Width,0,z);
  bpBelow:
    if BiDiMode=bdRightToLeft then
      fBack.Canvas.Draw(0,Height-z.Height-1,z)
    else
      fBack.Canvas.Draw(Width-z.Width,Height-z.Height-1,z);
end;
z.Free;
k:=0;
j:=0;
if GetPrevButtonVisible then
  begin
  case fButtonPos of
    bpLeft:
      if fButtonRot then
        begin
        k:=fLMargin+fButtonHeight div 2;
        j:=fTMargin+4;
        d:=adUp;
        e:=Bounds(fLMargin+1,fTMargin+3,fButtonHeight-2,GetArrowSize);
        end
      else
        begin
        k:=fLMargin+fButtonWidth div 2;
        j:=fTMargin+4;
        d:=adUp;
        e:=Bounds(fLMargin+1,fTMargin+3,fButtonWidth-2,GetArrowSize);
        end;
    bpRight:
      if fButtonRot then
        begin
        k:=Width-Integer(fRMargin)-Integer(fButtonHeight) div 2;
        j:=fTMargin+4;
        d:=adUp;
        e:=Bounds(Width-Integer(fRMargin+fButtonHeight),fTMargin+3,fButtonHeight-2,GetArrowSize);
        end
      else
        begin
        k:=Width-Integer(fRMargin)-Integer(fButtonWidth) div 2;
        j:=fTMargin+4;
        d:=adUp;
        e:=Bounds(Width-Integer(fRMargin+fButtonWidth),fTMargin+3,fButtonWidth-2,GetArrowSize);
        end;
    bpAbove:
      if BiDiMode=bdRightToLeft then
        begin
        k:=Width-Integer(fRMargin)-6;
        j:=fTMargin+fButtonHeight div 2;
        d:=adRight;
        e:=Bounds(Width-Integer(fRMargin)-4-GetArrowSize,fTMargin+1,GetArrowSize,fButtonHeight-2);
        end
      else
        begin
        k:=fLMargin+4;
        j:=fTMargin+fButtonHeight div 2;
        d:=adLeft;
        e:=Bounds(fLMargin+3,fTMargin+1,GetArrowSize,fButtonHeight-2);
        end;
    bpBelow:
      if BiDiMode=bdRightToLeft then
        begin
        k:=Width-Integer(fRMargin)-6;
        j:=Height-Integer(fBMargin)-Integer(fButtonHeight div 2);
        d:=adRight;
        e:=Bounds(Width-Integer(fRMargin)-4-GetArrowSize,Height-Integer(fBMargin+fButtonHeight),GetArrowSize,fButtonHeight-2);
        end
      else
        begin
        k:=fLMargin+4;
        j:=Height-Integer(fBMargin)-Integer(fButtonHeight div 2);
        d:=adLeft;
        e:=Bounds(fLMargin+3,Height-Integer(fBMargin+fButtonHeight),GetArrowSize,fButtonHeight-2);
        end;
  end;
  if fCustomArrow then
    begin
    t:=bsNormal;
    if fOnTab=MCoPTabPreviousButton then t:=bsOver;
    if (MDown) and (fOnTab=MCoPTabPreviousButton) then t:=bsPressed;
    if Assigned(fDrawArrow) then fDrawArrow(Self,fBack.Canvas,e,d,t);
    end
  else
    begin
    if fOnTab=MCoPTabPreviousButton then
      begin
      fBack.Canvas.Pen.Color:=fOverColor;
      fBack.Canvas.Brush.Color:=fOverColor;
      fBack.Canvas.FillRect(e);
      end;
    Arrow(k,j);
    end;
  end;
if GetNextButtonVisible then
  begin
  case fButtonPos of
    bpLeft:
      if fButtonRot then
        begin
        k:=fLMargin+fButtonHeight div 2;
        j:=Height-Integer(fBMargin)-6;
        d:=adDown;
        e:=Bounds(fLMargin+1,Height-Integer(fBMargin)-4-GetArrowSize,fButtonHeight-2,GetArrowSize);
        end
      else
        begin
        k:=fLMargin+fButtonWidth div 2;
        j:=Height-Integer(fBMargin)-6;
        d:=adDown;
        e:=Bounds(fLMargin+1,Height-Integer(fBMargin)-4-GetArrowSize,fButtonWidth-2,GetArrowSize);
        end;
    bpRight:
      if fButtonRot then
        begin
        k:=Width-Integer(fRMargin)-Integer(fButtonHeight) div 2;
        j:=Height-Integer(fBMargin)-6;
        d:=adDown;
        e:=Bounds(Width-Integer(fRMargin+fButtonHeight),Height-Integer(fBMargin)-4-GetArrowSize,fButtonHeight-2,GetArrowSize);
        end
      else
        begin
        k:=Width-Integer(fRMargin)-Integer(fButtonWidth) div 2;
        j:=Height-Integer(fBMargin)-6;
        d:=adDown;
        e:=Bounds(Width-Integer(fRMargin+fButtonWidth),Height-Integer(fBMargin)-4-GetArrowSize,fButtonWidth-2,GetArrowSize);
        end;
    bpAbove:
      if BiDiMode=bdRightToLeft then
        begin
        k:=fLMargin+4;
        j:=fTMargin+fButtonHeight div 2;
        d:=adLeft;
        e:=Bounds(fLMargin+3,fTMargin+1,GetArrowSize,fButtonHeight-2);
        end
      else
        begin
        k:=Width-Integer(fRMargin)-6;
        j:=fTMargin+fButtonHeight div 2;
        d:=adRight;
        e:=Bounds(Width-Integer(fRMargin)-4-GetArrowSize,fTMargin+1,GetArrowSize,fButtonHeight-2);
        end;
    bpBelow:
      if BiDiMode=bdRightToLeft then
        begin
        k:=fLMargin+4;
        j:=Height-Integer(fBMargin)-Integer(fButtonHeight div 2);
        d:=adLeft;
        e:=Bounds(fLMargin+3,Height-Integer(fBMargin+fButtonHeight),GetArrowSize,fButtonHeight-2);
        end
      else
        begin
        k:=Width-Integer(fRMargin)-6;
        j:=Height-Integer(fBMargin)-Integer(fButtonHeight div 2);
        d:=adRight;
        e:=Bounds(Width-Integer(fRMargin)-4-GetArrowSize,Height-Integer(fBMargin+fButtonHeight),GetArrowSize,fButtonHeight-2);
        end;
  end;
  if fCustomArrow then
    begin
    t:=bsNormal;
    if fOnTab=MCoPTabNextButton then t:=bsOver;
    if (MDown) and (fOnTab=MCoPTabNextButton) then t:=bsPressed;
    if Assigned(fDrawArrow) then fDrawArrow(Self,fBack.Canvas,e,d,t);
    end
  else
    begin
    if fOnTab=MCoPTabNextButton then
      begin
      fBack.Canvas.Pen.Color:=fOverColor;
      fBack.Canvas.Brush.Color:=fOverColor;
      fBack.Canvas.FillRect(e);
      end;
    Arrow(k,j);
    end;
  end;

SheetBack.Width:=fSheetRect.Right-fSheetRect.Left;
SheetBack.Height:=fSheetRect.Bottom-fSheetRect.Top;
SheetBack.Canvas.CopyRect(SheetBack.Canvas.ClipRect,fBack.Canvas,fSheetRect);
Paint;
end;

procedure TMCoPXPTabControl.Paint;
var r,g:HRGN;
begin
if not (Visible and Assigned(Parent)) then
  begin
  inherited Paint;
  Exit;
  end;
if Assigned(fActiveTab) then
  begin
  r:=CreateRectRGN(0,0,Width,Height);
  g:=CreateRectRGN(fSheetRect.Left,fSheetRect.Top,fSheetRect.Right,fSheetRect.Bottom);
  CombineRGN(r,r,g,RGN_DIFF);
  DeleteObject(g);
  SelectClipRGN(Canvas.Handle,r);
  Canvas.Draw(0,0,fBack);
  SelectClipRGN(Canvas.Handle,0);
  DeleteObject(r);
  end
else
  Canvas.Draw(0,0,fBack);
if Assigned(fPaint) then fPaint(Self);
end;

procedure TMCoPXPTabControl.Gradient(C1,C2:TColor;Ho:Boolean;var Bi:{$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap);
var
  c,m:Integer;
  r,g,b:Real;
  q,w,e:Byte;
begin
if Ho then m:=Bi.Height else m:=Bi.Width;
q:=GetRValue(C1);
w:=GetGValue(C1);
e:=GetBValue(C1);
r:=(GetRValue(C2)-q)/m;
g:=(GetGValue(C2)-w)/m;
b:=(GetBValue(C2)-e)/m;
for c:=0 to m-1 do
  begin
  if Ho then Bi.Canvas.MoveTo(0,c) else Bi.Canvas.MoveTo(c,0);
  Bi.Canvas.Pen.Color:=RGB(RealToByte(q+r*c),RealToByte(w+g*c),RealToByte(e+b*c));
  if Ho then Bi.Canvas.LineTo(Bi.Width,c) else Bi.Canvas.LineTo(c,Bi.Height);
  end;
end;

function TMCoPXPTabControl.RealToByte(R:Real):Byte;
begin
if R<=0 then Result:=0
else if R>=255 then Result:=255
else Result:=Round(R);
end;

procedure TMCoPXPTabControl.SetBackColor1(V:TColor);
begin
fBackColor1:=V;
fColorScheme:=csCustom;
Ref;
end;

procedure TMCoPXPTabControl.SetBackColor2(V:TColor);
begin
fBackColor2:=V;
fColorScheme:=csCustom;
Ref;
end;

procedure TMCoPXPTabControl.SetBackDir(V:TMCoPXPTabGradientDirection);
begin
fBackDir:=V;
Ref;
end;

procedure TMCoPXPTabControl.Resize;
var k:Integer;
begin
k:=0;
if (fButtonPos in [bpLeft,bpRight]) and (not fButtonRot) then k:=fButtonWidth;
if (fButtonPos in [bpLeft,bpRight]) and (fButtonRot) then k:=fButtonHeight;
if Integer(fRMargin)+Integer(fLMargin)+k+3>Width then Width:=Integer(fRMargin)+Integer(fLMargin)+k+3;
k:=0;
if fButtonPos in [bpAbove,bpBelow] then k:=fButtonHeight;
if Integer(fBMargin)+Integer(fTMargin)+k+3>Height then Height:=Integer(fBMargin)+Integer(fTMargin)+k+3;

if (fButtonRot and (fButtonPos in [bpLeft,bpRight])) and (Integer(fButtonWidth)>=Height-Integer(fTMargin)-Integer(fBMargin)-8) then fButtonWidth:=Height-Integer(fTMargin)-Integer(fBMargin)-9;
if (not (fButtonRot and (fButtonPos in [bpLeft,bpRight]))) and (Integer(fButtonWidth)>=Width-Integer(fLMargin)-Integer(fTMargin)-8) then fButtonWidth:=Width-Integer(fLMargin)-Integer(fTMargin)-9;
if (fButtonRot and (fButtonPos in [bpLeft,bpRight])) and (Integer(fButtonHeight)>=Width-Integer(fLMargin)-Integer(fTMargin)-8) then fButtonHeight:=Width-Integer(fLMargin)-Integer(fTMargin)-9;
if (not (fButtonRot and (fButtonPos in [bpLeft,bpRight]))) and (Integer(fButtonHeight)>=Height-Integer(fTMargin)-Integer(fBMargin)-8) then fButtonHeight:=Height-Integer(fTMargin)-Integer(fBMargin)-9;

inherited Resize;
CalSheetRect;
SetScroll(fScroll);
RefTabs;
end;

procedure TMCoPXPTabControl.SetButtonWidth(V:Cardinal);
begin
if V<10 then V:=10;
if (fButtonRot and (fButtonPos in [bpLeft,bpRight])) and (Integer(V)>=Height-Integer(fTMargin)-Integer(fBMargin)-8) then V:=Height-Integer(fTMargin)-Integer(fBMargin)-9;
if (not (fButtonRot and (fButtonPos in [bpLeft,bpRight]))) and (Integer(V)>=Width-Integer(fLMargin)-Integer(fTMargin)-8) then V:=Width-Integer(fLMargin)-Integer(fTMargin)-9;
if fTMR+fTML>Integer(V) then V:=fTMR+fTML;
fButtonWidth:=V;
CalSheetRect;
SetScroll(fScroll);
RefTabs;
end;

procedure TMCoPXPTabControl.SetButtonHeight(V:Cardinal);
begin
if V<10 then V:=10;
if (fButtonRot and (fButtonPos in [bpLeft,bpRight])) and (Integer(V)>=Width-Integer(fLMargin)-Integer(fTMargin)-8) then V:=Width-Integer(fLMargin)-Integer(fTMargin)-9;
if (not (fButtonRot and (fButtonPos in [bpLeft,bpRight]))) and (Integer(V)>=Height-Integer(fTMargin)-Integer(fBMargin)-8) then V:=Height-Integer(fTMargin)-Integer(fBMargin)-9;
if fTMB+fTMT>Integer(V) then V:=fTMB+fTMT;
fButtonHeight:=V;
CalSheetRect;
SetScroll(fScroll);
RefTabs;
end;

procedure TMCoPXPTabControl.SetLMargin(V:Cardinal);
var k:Integer;
begin
k:=0;
if (fButtonPos in [bpLeft,bpRight]) and (not fButtonRot) then k:=fButtonWidth;
if (fButtonPos in [bpLeft,bpRight]) and (fButtonRot) then k:=fButtonHeight;
if Integer(fRMargin)+Integer(V)+k+3>Width then V:=Width-Integer(fRMargin)-k-3;
fLMargin:=V;
CalSheetRect;
SetScroll(fScroll);
RefTabs;
end;

procedure TMCoPXPTabControl.SetRMargin(V:Cardinal);
var k:Integer;
begin
k:=0;
if (fButtonPos in [bpLeft,bpRight]) and (not fButtonRot) then k:=fButtonWidth;
if (fButtonPos in [bpLeft,bpRight]) and (fButtonRot) then k:=fButtonHeight;
if Integer(fLMargin)+Integer(V)+k+3>Width then V:=Width-Integer(fLMargin)-k-3;
fRMargin:=V;
CalSheetRect;
SetScroll(fScroll);
RefTabs;
end;

procedure TMCoPXPTabControl.SetTMargin(V:Cardinal);
var k:Integer;
begin
k:=0;
if fButtonPos in [bpAbove,bpBelow] then k:=fButtonHeight;
if Integer(fBMargin)+Integer(V)+k+3>Height then V:=Height-Integer(fBMargin)-k-3;
fTMargin:=V;
CalSheetRect;
SetScroll(fScroll);
RefTabs;
end;

procedure TMCoPXPTabControl.SetBMargin(V:Cardinal);
var k:Integer;
begin
k:=0;
if fButtonPos in [bpAbove,bpBelow] then k:=fButtonHeight;
if Integer(fTMargin)+Integer(V)+k+3>Height then V:=Height-Integer(fTMargin)-k-3;
fBMargin:=V;
CalSheetRect;
SetScroll(fScroll);
RefTabs;
end;

procedure TMCoPXPTabControl.SetButtonPos(V:TMCoPXPTabButtonsPosition);
begin
fButtonPos:=V;
Resize;
end;

procedure TMCoPXPTabControl.SetButtonRot(V:Boolean);
begin
fButtonRot:=V;
Resize;
end;

procedure TMCoPXPTabControl.SetBoxColor1(V:TColor);
begin
fBoxColor1:=V;
fColorScheme:=csCustom;
Ref;
RefTabs;
end;

procedure TMCoPXPTabControl.SetBoxColor2(V:TColor);
begin
fBoxColor2:=V;
fColorScheme:=csCustom;
Ref;
RefTabs;
end;

function TMCoPXPTabControl.BlendColors(Back,Fore:TColor;A:Byte):TColor;
var q,w,e,r,g,b:Byte;
begin
if a=0 then
 Result:=Back
else if a=255 then
 Result:=Fore
else
 begin
 r:=getrvalue(Fore);
 g:=getgvalue(Fore);
 b:=getbvalue(Fore);
 q:=getrvalue(Back);
 w:=getgvalue(Back);
 e:=getbvalue(Back);
 R:=(R*A+q*(255-A)) div 255;
 G:=(G*A+w*(255-A)) div 255;
 B:=(B*A+e*(255-A)) div 255;
 Result:=rgb(R,G,B);
 end;
end;

procedure TMCoPXPTabControl.Notification(AComponent:TComponent;Operation:TOperation);
begin
inherited Notification(AComponent,Operation);
if csDestroying in ComponentState then Exit;
if (Operation=opRemove) and (AComponent=fActiveTab) then
  RefActiveTab
else if (Operation=opRemove) and (AComponent is TMCoPXPTabSheet) then
  if (AComponent as TMCoPXPTabSheet).TabControl=Self then
    SetScroll(fScroll)
else if (Operation=opRemove) and (AComponent=fImages) then
  fImages:=nil;
end;

function TMCoPXPTabControl.GetTabCount:Cardinal;
begin
Result:=ControlCount;
end;

procedure TMCoPXPTabControl.SetActiveTab(V:TMCoPXPTabSheet);
begin
if fActiveTab<>V then
  begin
  if not CanChange then Exit;
  if V<>nil then
    begin
    if V.Parent<>Self then
      raise EMCoPXPTabError.Create('TabSheet dose not belongs to TabControl');
    if not ((V.TabVisible) or (csDesigning in ComponentState)) then
      raise EMCoPXPTabError.Create('Can not activate an invisible TabSheet');
    end;
  fActiveTab:=V;
  if V<>nil then RefScroll;
  Ref;
  RefTabs;
  if Assigned(fChange) then fChange(Self);
  end;
end;

function TMCoPXPTabControl.GetTab(Index:Cardinal):TMCoPXPTabSheet;
begin
Result:=TMCoPXPTabSheet(Controls[Index]);
end;

function TMCoPXPTabControl.GetTabIndex:Integer;
var k:Integer;
begin
Result:=-1;
for k:=0 to ControlCount-1 do
  if Controls[k]=fActiveTab then
    begin
    Result:=k;
    Break;
    end;
end;

procedure TMCoPXPTabControl.SetTabIndex(V:Integer);
begin
if (V<0) or (V>=ControlCount) then
  raise EMCoPXPTabError.CreateFmt('Tab index out of bounds (%d)',[V])
else
  SetActiveTab(TMCoPXPTabSheet(Controls[V]));
end;

procedure TMCoPXPTabControl.AdjustClientRect(var Rect:TRect);
begin
Rect:=fSheetRect;
inherited;
end;

procedure TMCoPXPTabControl.CalSheetRect;
var l,t,w,h:Integer;
begin
w:=Width-Integer(fLMargin+fRMargin);
h:=Height-Integer(fTMargin+fBMargin);
case fButtonPos of
  bpAbove,bpBelow:h:=h-Integer(fButtonHeight);
  bpLeft,bpRight:
    if fButtonRot then
      w:=w-Integer(fButtonHeight)
    else
      w:=w-Integer(fButtonWidth);
end;
l:=fLMargin;
t:=fTMargin;
case fButtonPos of
  bpAbove:t:=t+Integer(fButtonHeight);
  bpLeft:
    if fButtonRot then
      l:=l+Integer(fButtonHeight)
    else
      l:=l+Integer(fButtonWidth);
end;
fSheetRect:=Bounds(l+1,t+1,w-3,h-3);
end;

procedure TMCoPXPTabControl.RefTabs;
var k:Integer;
begin
for k:=0 to ControlCount-1 do
  begin
  TMCoPXPTabSheet(Controls[k]).Width:=fSheetRect.Right-fSheetRect.Left;
  TMCoPXPTabSheet(Controls[k]).Height:=fSheetRect.Bottom-fSheetRect.Top;
  TMCoPXPTabSheet(Controls[k]).Img.Picture.Assign(SheetBack);
  TMCoPXPTabSheet(Controls[k]).Img.SendToBack;
  TMCoPXPTabSheet(Controls[k]).Visible:=False;
  end;
if Assigned(fActiveTab) then fActiveTab.Show;
end;

procedure TMCoPXPTabControl.SetTML(V:Integer);
begin
if fTMR+V>Integer(fButtonWidth) then V:=Integer(fButtonWidth)-fTMR-1;
fTML:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetTMR(V:Integer);
begin
if fTML+V>Integer(fButtonWidth) then V:=Integer(fButtonWidth)-fTML-1;
fTMR:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetTMT(V:Integer);
begin
if fTMB+V>Integer(fButtonHeight) then V:=Integer(fButtonHeight)-fTMB-1;
fTMT:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetTMB(V:Integer);
begin
if fTMT+V>Integer(fButtonHeight) then V:=Integer(fButtonHeight)-fTMT-1;
fTMB:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetImages(V:TCustomImageList);
begin
if Assigned(fImages) then fImages.UnRegisterChanges(ImageChangeLink);
fImages:=V;
if fImages<>nil then
  begin
  fImages.RegisterChanges(ImageChangeLink);
  fImages.FreeNotification(Self);
  end;
Ref;
end;

procedure TMCoPXPTabControl.FlipBitmap(CW:Boolean;var B:{$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap);
var
  d:{$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap;
  x,y:Integer;
  f:PByteArray;
begin
d:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
d.Width:=b.Height;
d.Height:=b.Width;
b.PixelFormat:=pf24bit;
for y:=0 to b.Height-1 do
  begin
  f:=b.ScanLine[y];
  for x:=0 to b.Width-1 do
    if CW then
      d.Canvas.Pixels[b.Height-y-1,x]:=RGB(f^[x*3],f^[x*3+1],f^[x*3+2])
    else
      d.Canvas.Pixels[y,b.Width-x-1]:=RGB(f^[x*3],f^[x*3+1],f^[x*3+2]);
  end;
b.Assign(d);
d.Free;
end;

procedure TMCoPXPTabControl.SetImgT(V:Integer);
begin
fImgT:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetImgN(V:Integer);
begin
fImgN:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetScroll(V:Cardinal);
begin
if V>GetScrollMax then V:=GetScrollMax;
fScroll:=V;
Ref;
end;

function TMCoPXPTabControl.GetPrevButtonVisible:Boolean;
begin
Result:=fScroll<>0;
end;

function TMCoPXPTabControl.GetNextButtonVisible:Boolean;
begin
Result:=fScroll<GetScrollMax;
end;

function TMCoPXPTabControl.GetScrollMax:Cardinal;
var k,j,c:Integer;
begin
c:=0;
for j:=0 to ControlCount-1 do
  if TabVisible(TMCoPXPTabSheet(Controls[j])) then
    Inc(c);
j:=0;
if (fButtonRot) or (fButtonPos in [bpAbove,bpBelow]) then
  k:=fButtonWidth+1
else
  k:=fButtonHeight+1;
case fButtonPos of
  bpAbove,bpBelow:j:=Width-Integer(fLMargin)-Integer(fRMargin)-8;
  bpLeft,bpRight:j:=Height-Integer(fTMargin)-Integer(fBMargin)-8;
end;
if k*c-j<0 then
  Result:=0
else
  Result:=c-(j div k);
end;

procedure TMCoPXPTabControl.SetArrowColor(V:TColor);
begin
fArrowColor:=V;
fColorScheme:=csCustom;
Ref;
end;

procedure TMCoPXPTabControl.SetOverColor(V:TColor);
begin
fOverColor:=V;
fColorScheme:=csCustom;
Ref;
end;

procedure TMCoPXPTabControl.MouseUp(Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
begin
MDown:=False;
if TabDown=fOnTab then
  begin
  if TabDown>MCoPTabNone then SetTabIndex(TabDown);
  if TabDown=MCoPTabPreviousButton then SetScroll(fScroll-1);
  if TabDown=MCoPTabNextButton then SetScroll(fScroll+1);
  if TabDown=MCoPTabCloseButton then DoCloseTab;
  MousePos(LX,LY);
  end;
inherited MouseUp(Button,Shift,X,Y);
end;

procedure TMCoPXPTabControl.MouseDown(Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
begin
MDown:=Button=mbLeft;
if MDown then TabDown:=fOnTab else TabDown:=MCoPTabNone;
if (TabDown<>MCoPTabNone) and (TabDown=GetTabIndex) and (CanFocus) then SetFocus;
Ref;
inherited MouseDown(Button,Shift,X,Y);
end;

procedure TMCoPXPTabControl.MouseMove(Shift:TShiftState;X,Y:Integer);
begin
LX:=X;
LY:=Y;
MousePos(X,Y);
inherited MouseMove(Shift,X,Y);
end;

procedure TMCoPXPTabControl.MousePos(x,y:Integer);
var
  r,n,m,i:Integer;
label
  a;
begin
r:=MCoPTabNone;
case fButtonPos of
  bpAbove:
    begin
    if (y<Integer(fTMargin)) or (y>Integer(fTMargin+fButtonHeight)) then goto a;
    if GetPrevButtonVisible then i:=GetArrowSize+4 else i:=4;
    if GetNextButtonVisible then n:=GetArrowSize+4 else n:=4;
    if BiDiMode=bdRightToLeft then
      begin
      m:=n;
      n:=i;
      i:=m;
      end;
    if (x>Integer(fLMargin)+4) and (x<Integer(fLMargin)+GetArrowSize+4) and (GetPrevButtonVisible) and (BiDiMode<>bdRightToLeft) then r:=MCoPTabPreviousButton;
    if (x>Width-Integer(fRMargin)-GetArrowSize-4) and (x<Width-Integer(fRMargin)-4) and (GetNextButtonVisible) and (BiDiMode<>bdRightToLeft) then r:=MCoPTabNextButton;
    if (x>Integer(fLMargin)+4) and (x<Integer(fLMargin)+GetArrowSize+4) and (GetNextButtonVisible) and (BiDiMode=bdRightToLeft) then r:=MCoPTabNextButton;
    if (x>Width-Integer(fRMargin)-GetArrowSize-4) and (x<Width-Integer(fRMargin)-4) and (GetPrevButtonVisible) and (BiDiMode=bdRightToLeft) then r:=MCoPTabPreviousButton;
    if (x<Integer(fLMargin)+i) or (x>Width-Integer(fRMargin)-n) then goto a;
    if BiDiMode=bdRightToLeft then
      r:=(Width-x-Integer(fRMargin)-i) div Integer(fButtonWidth)+Integer(fScroll)
    else
      r:=(x-Integer(fLMargin)-i) div Integer(fButtonWidth)+Integer(fScroll);
    end;
  bpBelow:
    begin
    if (y<Height-Integer(fBMargin+fButtonHeight)) or (y>Height-Integer(fBMargin)) then goto a;
    if GetPrevButtonVisible then i:=GetArrowSize+4 else i:=4;
    if GetNextButtonVisible then n:=GetArrowSize+4 else n:=4;
    if BiDiMode=bdRightToLeft then
      begin
      m:=n;
      n:=i;
      i:=m;
      end;
    if (x>Integer(fLMargin)+4) and (x<Integer(fLMargin)+GetArrowSize+4) and (GetPrevButtonVisible) and (BiDiMode<>bdRightToLeft) then r:=MCoPTabPreviousButton;
    if (x>Width-Integer(fRMargin)-GetArrowSize-4) and (x<Width-Integer(fRMargin)-4) and (GetNextButtonVisible) and (BiDiMode<>bdRightToLeft) then r:=MCoPTabNextButton;
    if (x>Integer(fLMargin)+4) and (x<Integer(fLMargin)+GetArrowSize+4) and (GetNextButtonVisible) and (BiDiMode=bdRightToLeft) then r:=MCoPTabNextButton;
    if (x>Width-Integer(fRMargin)-GetArrowSize-4) and (x<Width-Integer(fRMargin)-4) and (GetPrevButtonVisible) and (BiDiMode=bdRightToLeft) then r:=MCoPTabPreviousButton;
    if (x<Integer(fLMargin)+i) or (x>Width-Integer(fRMargin)-n) then goto a;
    if BiDiMode=bdRightToLeft then
      r:=(Width-x-Integer(fLMargin)-i) div Integer(fButtonWidth)+Integer(fScroll)
    else
      r:=(x-Integer(fLMargin)-i) div Integer(fButtonWidth)+Integer(fScroll);
    end;
  bpLeft:
    begin
    if fButtonRot then
      begin
      i:=fButtonHeight;
      m:=fButtonWidth;
      end
    else
      begin
      m:=fButtonHeight;
      i:=fButtonWidth;
      end;
    if (x<Integer(fLMargin)) or (x>Integer(fLMargin)+i) then goto a;
    if GetPrevButtonVisible then i:=GetArrowSize+4 else i:=4;
    if GetNextButtonVisible then n:=GetArrowSize+4 else n:=4;
    if (y>Integer(fTMargin)+4) and (y<Integer(fTMargin)+GetArrowSize+4) and (GetPrevButtonVisible) then r:=MCoPTabPreviousButton;
    if (y>Height-Integer(fBMargin)-GetArrowSize-4) and (y<Height-Integer(fBMargin)-4) and (GetNextButtonVisible) then r:=MCoPTabNextButton;
    if (y<Integer(fTMargin)+i) or (y>Height-Integer(fBMargin)-n) then goto a;
    r:=(y-Integer(fTMargin)-i) div m+Integer(fScroll);
    end;
  bpRight:
    begin
    if fButtonRot then
      begin
      i:=fButtonHeight;
      m:=fButtonWidth;
      end
    else
      begin
      m:=fButtonHeight;
      i:=fButtonWidth;
      end;
    if (x<Width-Integer(fRMargin)-i) or (x>Width-Integer(fRMargin)) then goto a;
    if GetPrevButtonVisible then i:=GetArrowSize+4 else i:=4;
    if GetNextButtonVisible then n:=GetArrowSize+4 else n:=4;
    if (y>Integer(fTMargin)+4) and (y<Integer(fTMargin)+GetArrowSize+4) and (GetPrevButtonVisible) then r:=MCoPTabPreviousButton;
    if (y>Height-Integer(fBMargin)-GetArrowSize-4) and (y<Height-Integer(fBMargin)-4) and (GetNextButtonVisible) then r:=MCoPTabNextButton;
    if (y<Integer(fTMargin)+i) or (y>Height-Integer(fBMargin)-n) then goto a;
    r:=(y-Integer(fTMargin)-i) div m+Integer(fScroll);
    end;
end;
a:
if r>=ControlCount then r:=MCoPTabNone;
if r>=0 then
  for i:=0 to ControlCount-1 do
    begin
    if TabVisible(TMCoPXPTabSheet(Controls[i])) then Dec(r);
    if r=-1 then
      begin
      r:=i;
      Break;
      end;
    if (i=ControlCount-1) and (r<>-1) then r:=MCoPTabNone;
    end;
if r>ControlCount then r:=MCoPTabNone;
if (PtInRect(CloseButtonRect,Point(x,y))) and (fClose) then r:=MCoPTabCloseButton;
if r<>fOnTab then
  begin
  fOnTab:=r;
  Ref;
  end;
end;

procedure TMCoPXPTabControl.RefActiveTab;
var
  r:Boolean;
  i:Integer;
begin
fActiveTab.Visible:=False;
r:=False;
i:=GetTabIndex+1;
while i<ControlCount do
  begin
  if TabVisible(TMCoPXPTabSheet(Controls[i])) then
    begin
    SetTabIndex(i);
    r:=True;
    Break;
    end;
  Inc(i);
  end;
if r then Exit;
i:=GetTabIndex-1;
while i>=0 do
  begin
  if TabVisible(TMCoPXPTabSheet(Controls[i])) then
    begin
    SetTabIndex(i);
    r:=True;
    Break;
    end;
  Dec(i);
  end;
if r then Exit;
fActiveTab:=nil;
if Assigned(fChange) then fChange(Self);
SetScroll(fScroll);
RefTabs;
end;

procedure TMCoPXPTabControl.SetCustomArrow(V:Boolean);
begin
fCustomArrow:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetDrawArrow(V:TMCoPXPTabOnDrawArrow);
begin
fDrawArrow:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetCloseColor(V:TColor);
begin
fCloseColor:=V;
fColorScheme:=csCustom;
Ref;
end;

procedure TMCoPXPTabControl.SetClose(V:Boolean);
begin
fClose:=V;
ClsBtn.Visible:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetClsT(V:Integer);
begin
fClsT:=V;
if fClose then Ref;
end;

procedure TMCoPXPTabControl.SetClsN(V:Integer);
begin
fClsN:=V;
if fClose then Ref;
end;

procedure TMCoPXPTabControl.SetCustomClose(V:Boolean);
begin
fCustomClose:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetDrawClose(V:TMCoPXPTabOnDrawCloseButton);
begin
fDrawClose:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetCloseSize(V:TMCoPXPTabOnMeasureCloseButton);
begin
fCloseSize:=V;
Ref;
end;

procedure TMCoPXPTabControl.SetArrowSize(V:TMCoPXPTabOnMeasureArrow);
begin
fArrowSize:=V;
Ref;
end;

function TMCoPXPTabControl.GetArrowSize:Integer;
begin
Result:=10;
if not fCustomArrow then Exit;
if not Assigned(fArrowSize) then Exit;
fArrowSize(Self,Result);
Result:=Abs(Result);
end;

procedure TMCoPXPTabControl.DoCloseTab;
begin
if Assigned(fActiveTab.OnClose) then fActiveTab.OnClose(Self);
if Assigned(fCloseTab) then fCloseTab(Self,fActiveTab,GetTabIndex);
end;

procedure TMCoPXPTabControl.SetColorScheme(V:TMCoPXPTabColorScheme);
begin
fColorScheme:=V;
case fColorScheme of
  csWindows:
    begin
    fBackColor1:=ColorToRGB(clBtnFace);
    fBackColor2:=ColorToRGB(clBtnShadow);
    fBoxColor1:=ColorToRGB(clBtnHighlight);
    fBoxColor2:=ColorToRGB(clBtnFace);
    fArrowColor:=ColorToRGB(clBtnText);
    fOverColor:=ColorToRGB(clHighlight);
    fCloseColor:=ColorToRGB(clBtnText);
    fTabLine:=ColorToRGB(cl3DDkShadow);
    end;
  csBlue:
    begin
    fBackColor1:=$00E5AF8A;
    fBackColor2:=$00E5AF8A;
    fBoxColor1:=$00F8EDE3;
    fBoxColor2:=$00F1D8C6;
    fArrowColor:=$00000000;
    fOverColor:=$00F2E6CC;
    fCloseColor:=$00000000;
    fTabLine:=$00E5AF8A;
    end;
  csOlive:
    begin
    fBackColor1:=$008FC2B4;
    fBackColor2:=$0091C4B5;
    fBoxColor1:=$00CFF0EA;
    fBoxColor2:=$00CFF0EA;
    fArrowColor:=$00000000;
    fOverColor:=$00D8F3E8;
    fCloseColor:=$00000000;
    fTabLine:=$008FC2B4;
    end;
  csSilver:
    begin
    fBackColor1:=$00E6D8D8;
    fBackColor2:=$00E6D8D8;
    fBoxColor1:=$00F7F3F3;
    fBoxColor2:=$00E6D8D8;
    fArrowColor:=$00000000;
    fOverColor:=$00B7AAAC;
    fCloseColor:=$00000000;
    fTabLine:=$00D2B9B9;
    end;
  csSky:
    begin
    fBackColor1:=$00FAD9C3;
    fBackColor2:=$00E4AE88;
    fBoxColor1:=$00FDEADA;
    fBoxColor2:=$00FADAC4;
    fArrowColor:=$00000000;
    fOverColor:=$00FDF1D9;
    fCloseColor:=$00000000;
    fTabLine:=$00F5B487;
    end;
  csBlack:
    begin
    fBackColor1:=$00535353;
    fBackColor2:=$003A3A3A;
    fBoxColor1:=$00CFC6C1;
    fBoxColor2:=$00DAD8D6;
    fArrowColor:=$00E1DFD8;
    fOverColor:=$00A8A8A8;
    fCloseColor:=$00000000;
    fTabLine:=$00535353;
    end;
  csOrange:
    begin
    fBackColor1:=$0066CCFF;
    fBackColor2:=$00029AFF;
    fBoxColor1:=$009FE9FB;
    fBoxColor2:=$009FE9FB;
    fArrowColor:=$00000000;
    fOverColor:=$0070E1FE;
    fCloseColor:=$00000000;
    fTabLine:=$0000A4F2;
    end;
  csPurple:
    begin
    fBackColor1:=$00FF33FF;
    fBackColor2:=$00FF99FF;
    fBoxColor1:=$00FFD7FF;
    fBoxColor2:=$00FFAAFF;
    fArrowColor:=$00480000;
    fOverColor:=$009800AF;
    fCloseColor:=$00000000;
    fTabLine:=$00FF33FF;
    end;
  csGreen:
    begin
    fBackColor1:=$002B8034;
    fBackColor2:=$004DC45A;
    fBoxColor1:=$0042C050;
    fBoxColor2:=$0089CD90;
    fArrowColor:=$00000000;
    fOverColor:=$00A4D2AE;
    fCloseColor:=$00000000;
    fTabLine:=$002B8034;
    end;
  csRed:
    begin
    fBackColor1:=$00262684;
    fBackColor2:=$004646CA;
    fBoxColor1:=$00DCDCF5;
    fBoxColor2:=$009797E1;
    fArrowColor:=$00FFFFFF;
    fOverColor:=$006D7CE0;
    fCloseColor:=$00000000;
    fTabLine:=$00262684;
    end;
end;
Ref;
RefTabs;
end;

function TMCoPXPTabControl.StoreColor:Boolean;
begin
Result:=fColorScheme=csCustom;
end;

procedure TMCoPXPTabControl.CMMouseLeave(var Message:TMessage);
begin
MousePos(-1,-1);
inherited;
end;

function TMCoPXPTabControl.GetHomepage:String;
begin
Result:='http://www.dihav.com/';
end;

procedure TMCoPXPTabControl.SetTabLine(V:TColor);
begin
fTabLine:=V;
fColorScheme:=csCustom;
Ref;
end;

procedure TMCoPXPTabControl.WMSetFocus(var Message: TWMSetFocus);
begin
inherited;
Ref;
end;

procedure TMCoPXPTabControl.WMKillFocus(var Message: TWMSetFocus);
begin
inherited;
Ref;
end;

function TMCoPXPTabControl.TabVisible(Tab:TMCoPXPTabSheet):Boolean;
begin
if csDesigning in ComponentState then
  Result:=True
else
  Result:=Tab.TabVisible;
end;

procedure TMCoPXPTabControl.CNKeyDown(var Message:TWMKeyDown);
var
  i,j:Integer;
  s:TShiftState;
begin
s:=KeyDataToShiftState(Message.KeyData);
if ((ssShift in s) or (ssCtrl in s) or (ssAlt in s)) then
  begin
  inherited;
  Exit;
  end;
i:=GetTabIndex;
j:=0;
case Message.CharCode of
  VK_LEFT:
    if (fButtonPos in [bpAbove,bpBelow]) then
      begin
      if (BiDiMode=bdRightToLeft) and (i<Integer(ControlCount)-1) then j:=+1;
      if (BiDiMode<>bdRightToLeft) and (i>0) then j:=-1;
      end;
  VK_RIGHT:
    if (fButtonPos in [bpAbove,bpBelow]) then
      begin
      if (BiDiMode=bdRightToLeft) and (i>0) then j:=-1;
      if (BiDiMode<>bdRightToLeft) and (i<Integer(ControlCount)-1) then j:=+1;
      end;
  VK_UP:
    if (fButtonPos in [bpLeft,bpRight]) and (i>0) then j:=-1;
  VK_DOWN:
    if (fButtonPos in [bpLeft,bpRight]) and (i<Integer(ControlCount)-1) then j:=+1;
end;
if j<>0 then
  while not TabVisible(TMCoPXPTabSheet(Controls[i+j])) do
    begin
    j:=j+(j div Abs(j));
    if (j+i<0) or (j+i>=ControlCount) then
      begin
      j:=0;
      Break;
      end;
    end;
if j<>0 then SetTabIndex(i+j) else inherited;
end;

procedure TMCoPXPTabControl.RefScroll;
var c:Integer;
begin
c:=GetVisibleTabsCount-GetScrollMax;
if fActiveTab.ShowIndex<Integer(fScroll) then SetScroll(fActiveTab.ShowIndex);
if fActiveTab.ShowIndex>=Integer(fScroll)+c then SetScroll(fActiveTab.ShowIndex-c+1);
end;

function TMCoPXPTabControl.CanChange:Boolean;
begin
Result:=True;
if csDesigning in ComponentState then Exit;
if Assigned(fChanging) then fChanging(Self,Result);
end;

procedure TMCoPXPTabControl.CMDialogChar(var Message:TCMDialogChar);
var i:Integer;
begin
for i:=0 to ControlCount-1 do
  if IsAccel(Message.CharCode,GetTab(i).Caption) and (GetTab(i).TabVisible) and (CanFocus) then
    begin
    Message.Result:=1;
    SetTabIndex(i);
    Exit;
    end;
inherited;
end;

procedure TMCoPXPTabControl.TCM_ADJUSTRECT(var Message:TMessage);
var
  l,r,t,b,i:Integer;
  p:PRect;
begin
l:=fLMargin+1;
t:=fTMargin+1;
r:=fRMargin+1;
b:=fBMargin+1;
if fButtonRot then i:=fButtonHeight else i:=fButtonWidth;
case fButtonPos of
  bpAbove:Inc(t,fButtonHeight);
  bpBelow:Inc(b,fButtonHeight);
  bpLeft:Inc(l,i);
  bpRight:Inc(r,i);
end;
p:=Ptr(Message.LParam);
if Boolean(Message.WParam) then i:=-1 else i:=1;
Inc(p^.Left,l*i);
Inc(p^.Top,t*i);
Dec(p^.Right,r*i);
Dec(p^.Bottom,b*i);
end;

procedure TMCoPXPTabControl.TCM_DELETEALLITEMS(var Message:TMessage);
var
  k:Integer;
  b:Boolean;
begin
b:=fActiveTab<>nil;;
for k:=ControlCount-1 downto 0 do
  begin
  GetTab(k).TabControl:=nil;
  Controls[k].Destroy;
  end;
fActiveTab:=nil;
Message.Result:=Integer(True);
if b and Assigned(fChange) then fChange(Self);
Ref;
RefTabs;
end;

procedure TMCoPXPTabControl.TCM_DELETEITEM(var Message:TMessage);
begin
if (Message.WParam<0) or (Message.WParam>=ControlCount) then
  Message.Result:=Integer(False)
else
  begin
  if GetTabIndex=Message.WParam then RefActiveTab;
  GetTab(Message.WParam).TabControl:=nil;
  Controls[Message.WParam].Free;
  Message.Result:=Integer(True);
  Ref;
  end;
end;

procedure TMCoPXPTabControl.TCM_GETCURFOCUS(var Message:TMessage);
begin
Message.Result:=GetTabIndex;
end;

procedure TMCoPXPTabControl.TCM_GETCURSEL(var Message:TMessage);
begin
Message.Result:=GetTabIndex;
end;

procedure TMCoPXPTabControl.TCM_GETIMAGELIST(var Message:TMessage);
begin
if Assigned(fImages) then
  Message.Result:=fImages.Handle
else
  Message.Result:=0;
end;

procedure TMCoPXPTabControl.TCM_GETITEM(var Message:TMessage);
var
  a:PTCItem;
  t:TMCoPXPTabSheet;
  k:Integer;
begin
if (Message.WParam<0) or (Message.WParam>=ControlCount) then
  Message.Result:=Integer(False)
else
  begin
  a:=Ptr(Message.LParam);
  t:=GetTab(Message.WParam);
  if (a^.mask and TCIF_TEXT)=TCIF_TEXT then
    begin
    k:=Length(t.Caption)+1;
    if k>a.cchTextMax then k:=a^.cchTextMax;
    CopyMemory(a^.pszText,PChar(t.Caption),k);
    end;
  if (a^.mask and TCIF_IMAGE)=TCIF_IMAGE then
    a^.iImage:=t.ImageIndex;
  Message.Result:=Integer(True);
  end;
end;

procedure TMCoPXPTabControl.TCM_GETITEMCOUNT(var Message:TMessage);
begin
Message.Result:=ControlCount;
end;

procedure TMCoPXPTabControl.TCM_GETITEMRECT(var Message:TMessage);
var
  a:PRect;
  t:TRect;
begin
if (Message.WParam<0) or (Message.WParam>ControlCount-1) then
  Message.Result:=Integer(False)
else
  begin
  a:=Ptr(Message.LParam);
  t:=GetRect(Message.WParam);
  CopyMemory(a,@t,SizeOf(t));
  Message.Result:=Integer(True);
  end;
end;

procedure TMCoPXPTabControl.TCM_GETROWCOUNT(var Message:TMessage);
begin
Message.Result:=1;
end;

procedure TMCoPXPTabControl.TCM_GETTOOLTIPS(var Message:TMessage);
begin
Message.Result:=0;
end;

procedure TMCoPXPTabControl.TCM_HITTEST(var Message:TMessage);
var a:PHDHitTestInfo;
begin
a:=Ptr(Message.LParam);
MousePos(a^.Point.X,a^.Point.Y);
if fOnTab>=0 then
  a^.Flags:=TCHT_ONITEM
else
  a^.Flags:=TCHT_NOWHERE;
Message.Result:=fOnTab;
MousePos(LX,LY);
end;

procedure TMCoPXPTabControl.TCM_INSERTITEM(var Message:TMessage);
var
  a:PTCItem;
  t:TMCoPXPTabSheet;
  b:array [0..255] of Char;
begin
if (Message.WParam<0) or (Message.WParam>ControlCount) then
  Message.Result:=-1
else
  begin
  a:=Ptr(Message.LParam);
  t:=TMCoPXPTabSheet.Create(Self);
  t.Parent:=Self;
  t.TabControl:=Self;
  if (a^.mask and TCIF_TEXT)=TCIF_TEXT then
    begin
    FillChar(b,256,0);
    CopyMemory(@b,a^.pszText,a^.cchTextMax);
    t.Caption:=b;
    end;
  if (a^.mask and TCIF_IMAGE)=TCIF_IMAGE then
    t.ImageIndex:=a^.iImage;
  t.PageIndex:=Message.WParam;
  Message.Result:=t.PageIndex;
  end;
end;

procedure TMCoPXPTabControl.TCM_REMOVEIMAGE(var Message:TMessage);
begin
if not Assigned(fImages) then Exit;
fImages.Delete(Message.WParam);
Ref;
end;

procedure TMCoPXPTabControl.TCM_SETCURFOCUS(var Message:TMessage);
begin
if Message.WParam<0 then
  SetActiveTab(nil)
else if Message.WParam<ControlCount then
  SetTabIndex(Message.WParam);
end;

procedure TMCoPXPTabControl.TCM_SETCURSEL(var Message:TMessage);
begin
if Message.WParam<0 then
  begin
  Message.Result:=GetTabIndex;
  SetActiveTab(nil);
  end
else if Message.WParam>ControlCount-1 then
  Message.Result:=-1
else
  begin
  Message.Result:=GetTabIndex;
  SetTabIndex(Message.WParam);
  end;
end;

procedure TMCoPXPTabControl.TCM_SETIMAGELIST(var Message:TMessage);
begin
if Assigned(fImages) then
  Message.Result:=fImages.Handle
else
  Message.Result:=0;
fImages:=TImageList.Create(Self);
fImages.Handle:=Message.lParam;
Ref;
end;

procedure TMCoPXPTabControl.TCM_SETITEM(var Message:TMessage);
var
  a:PTCItem;
  t:TMCoPXPTabSheet;
  b:array [0..255] of Char;
begin
if (Message.WParam<0) or (Message.WParam>ControlCount-1) then
  Message.Result:=Integer(False)
else
  begin
  a:=Ptr(Message.LParam);
  t:=GetTab(Message.WParam);
  if (a^.mask and LVIF_TEXT)=LVIF_TEXT then
    begin
    FillChar(b,256,0);
    CopyMemory(@b,a^.pszText,a^.cchTextMax);
    t.Caption:=b;
    end;
  if (a^.mask and LVIF_IMAGE)=LVIF_IMAGE then
    t.ImageIndex:=a^.iImage;
  Message.Result:=Integer(True);
  end;
end;

procedure TMCoPXPTabControl.TCM_SETITEMEXTRA(var Message:TMessage);
begin
Message.Result:=Integer(False);
end;

procedure TMCoPXPTabControl.TCM_SETITEMSIZE(var Message:TMessage);
begin
Message.Result:=MakeLParam(fButtonWidth,fButtonHeight);
fButtonWidth:=Message.LParamLo;
fButtonHeight:=Message.LParamHi;
Resize;
end;

procedure TMCoPXPTabControl.TCM_SETPADDING(var Message:TMessage);
begin
//
end;

procedure TMCoPXPTabControl.TCM_SETTOOLTIPS(var Message:TMessage);
begin
//
end;

procedure TMCoPXPTabControl.TCM_SETMINTABWIDTH(var Message:TMessage);
begin
Message.Result:=0;
end;

procedure TMCoPXPTabControl.TCM_DESELECTALL(var Message:TMessage);
begin
//
end;

procedure TMCoPXPTabControl.TCM_HIGHLIGHTITEM(var Message:TMessage);
begin
//
end;

procedure TMCoPXPTabControl.TCM_SETEXTENDEDSTYLE(var Message:TMessage);
begin
//
end;

procedure TMCoPXPTabControl.TCM_GETEXTENDEDSTYLE(var Message:TMessage);
begin
//
end;

procedure TMCoPXPTabControl.TCM_SETUNICODEFORMAT(var Message:TMessage);
begin
//
end;

procedure TMCoPXPTabControl.TCM_GETUNICODEFORMAT(var Message:TMessage);
begin
//
end;

procedure TMCoPXPTabControl.ImageListChange(Sender:TObject);
begin
Ref;
end;

function TMCoPXPTabControl.GetVisibleTabsCount:Cardinal;
var k:Integer;
begin
if csDesigning in ComponentState then
  begin
  Result:=ControlCount;
  Exit;
  end;
Result:=0;
for k:=0 to ControlCount-1 do
  if GetTab(k).TabVisible then
    Inc(Result);
end;

procedure TMCoPXPTabControl.SetRTL(V:Boolean);
begin
fRTL:=V;
Ref;
end;

function TMCoPXPTabControl.GetRect(Index:Cardinal):TRect;
var x,y,k,j,w,h,i:Integer;
begin
w:=fButtonWidth;
h:=fButtonHeight;
x:=0;
y:=0;
k:=0;
j:=0;
case fButtonPos of
  bpAbove:
    if BiDiMode=bdRightToLeft then
      begin
      k:=-w;
      j:=0;
      x:=Width-Integer(fRMargin)-4+k;
      y:=fTMargin;
      end
    else
      begin
      k:=w;
      j:=0;
      x:=fLMargin+4;
      y:=fTMargin;
      end;
  bpBelow:
    if BiDiMode=bdRightToLeft then
      begin
      k:=-w;
      j:=0;
      x:=Width-Integer(fRMargin)-4+k;
      y:=Height-Integer(fBMargin+fButtonHeight);
      end
    else
      begin
      k:=w;
      j:=0;
      x:=fLMargin+4;
      y:=Height-Integer(fBMargin+fButtonHeight);
      end;
  bpLeft:
    begin
    if fButtonRot then
      begin
      w:=fButtonHeight;
      h:=fButtonWidth;
      end;
    k:=0;
    j:=h;
    x:=fLMargin;
    y:=fTMargin+4;
    end;
  bpRight:
    begin
    if fButtonRot then
      begin
      w:=fButtonHeight;
      h:=fButtonWidth;
      end;
    k:=0;
    j:=h;
    x:=Width-Integer(fRMargin)-w;
    y:=fTMargin+4;
    end;
end;
i:=GetTab(Index).ShowIndex-Integer(fScroll);
x:=x+i*k;
y:=y+i*j;
Result:=Bounds(x,y,w,h);
if Result.Left<Integer(fLMargin) then Result.Left:=Integer(fLMargin);
if Result.Right<Integer(fLMargin) then Result.Right:=Integer(fLMargin);
if Result.Left>Width-Integer(fRMargin) then Result.Left:=Width-Integer(fRMargin);
if Result.Right>Width-Integer(fRMargin) then Result.Right:=Width-Integer(fRMargin);
if Result.Top<Integer(fTMargin) then Result.Top:=Integer(fTMargin);
if Result.Bottom<Integer(fTMargin) then Result.Bottom:=Integer(fTMargin);
if Result.Top>Height-Integer(fBMargin) then Result.Top:=Height-Integer(fBMargin);
if Result.Bottom>Height-Integer(fBMargin) then Result.Bottom:=Height-Integer(fBMargin);
end;

procedure TMCoPXPTabControl.RedrawControl;
begin
Ref;
RefTabs;
end;

procedure TMCoPXPTabControl.Assign(Source:TPersistent);
var s:TMCoPXPTabControl;
begin
if Source is TMCoPXPTabControl then
  begin
  s:=Source as TMCoPXPTabControl;
  ColorScheme:=s.ColorScheme;
  if ColorScheme=csCustom then
    begin
    ArrowColor:=s.ArrowColor;
    BackColor1:=s.BackColor1;
    BackColor2:=s.BackColor2;
    BoxColor1:=s.BoxColor1;
    BoxColor2:=s.BoxColor2;
    CloseButtonColor:=s.CloseButtonColor;
    FocusLineColor:=s.FocusLineColor;
    OverColor:=s.OverColor;
    end;
  BackGroundGradientDirection:=s.BackgroundGradientDirection;
  CloseButtonNormalPosition:=s.CloseButtonNormalPosition;
  CloseButtonTangentialPosition:=s.CloseButtonTangentialPosition;
  CustomArrow:=s.CustomArrow;
  CustomClose:=s.CustomClose;
  DrawCloseButton:=s.DrawCloseButton;
  ImageNormalPosition:=s.ImageNormalPosition;
  ImageTangentialPosition:=s.ImageTangentialPosition;
  Images:=s.Images;
  MarginBottom:=s.MarginBottom;
  MarginLeft:=s.MarginLeft;
  MarginRight:=s.MarginRight;
  MarginTop:=s.MarginTop;
  RotateTabs:=s.RotateTabs;
  RTLReading:=s.RTLReading;
  TabsHeight:=s.TabsHeight;
  TabsPosition:=s.TabsPosition;
  TabsWidth:=s.TabsWidth;
  TextMarginBottom:=s.TextMarginBottom;
  TextMarginLeft:=s.TextMarginLeft;
  TextMarginRight:=s.TextMarginRight;
  TextMarginTop:=s.TextMarginTop;
  end
else
  inherited Assign(Source);
end;

procedure TMCoPXPTabControl.AssignTo(Dest:TPersistent);
var d:TMCoPXPTabControl;
begin
if Dest is TMCoPXPTabControl then
  begin
  d:=Dest as TMCoPXPTabControl;
  d.ColorScheme:=ColorScheme;
  if d.ColorScheme=csCustom then
    begin
    d.ArrowColor:=ArrowColor;
    d.BackColor1:=BackColor1;
    d.BackColor2:=BackColor2;
    d.BoxColor1:=BoxColor1;
    d.BoxColor2:=BoxColor2;
    d.CloseButtonColor:=CloseButtonColor;
    d.FocusLineColor:=FocusLineColor;
    d.OverColor:=OverColor;
    end;
  d.BackGroundGradientDirection:=BackgroundGradientDirection;
  d.CloseButtonNormalPosition:=CloseButtonNormalPosition;
  d.CloseButtonTangentialPosition:=CloseButtonTangentialPosition;
  d.CustomArrow:=CustomArrow;
  d.CustomClose:=CustomClose;
  d.DrawCloseButton:=DrawCloseButton;
  d.ImageNormalPosition:=ImageNormalPosition;
  d.ImageTangentialPosition:=ImageTangentialPosition;
  d.Images:=Images;
  d.MarginBottom:=MarginBottom;
  d.MarginLeft:=MarginLeft;
  d.MarginRight:=MarginRight;
  d.MarginTop:=MarginTop;
  d.RotateTabs:=RotateTabs;
  d.RTLReading:=RTLReading;
  d.TabsHeight:=TabsHeight;
  d.TabsPosition:=TabsPosition;
  d.TabsWidth:=TabsWidth;
  d.TextMarginBottom:=TextMarginBottom;
  d.TextMarginLeft:=TextMarginLeft;
  d.TextMarginRight:=TextMarginRight;
  d.TextMarginTop:=TextMarginTop;
  end
else
  inherited AssignTo(Dest);
end;

procedure TMCoPXPTabControl.CMFontChanged(var Message:TMessage);
begin
Ref;
inherited;
end;

function TMCoPXPTabControl.GetControlCount:Integer;
begin
Result:=inherited ControlCount;
Dec(Result);
end;

function TMCoPXPTabControl.GetControl(Index:Integer):TControl;
begin
ClsBtn.SendToBack;
Result:=inherited Controls[Index+1];
end;

procedure TMCoPXPTabControl.ClsBtnClick(Sender:TObject);
begin
if fClose then DoCloseTab;
end;

procedure TMCoPXPTabControl.ClsBtnEnter(Sender:TObject);
begin
ClsFoc:=True;
if fClose then Ref;
end;

procedure TMCoPXPTabControl.ClsBtnExit(Sender:TObject);
begin
ClsFoc:=False;
if fClose then Ref;
end;

procedure TMCoPXPTabControl.EnterActivationCode(Name,Code:String);
begin
//
end;

//Sheet

constructor TMCoPXPTabSheet.Create(AOwner:TComponent);
begin
inherited Create(AOwner);
Align:=alClient;
ControlStyle:=ControlStyle+[csAcceptsControls,csNoDesignVisible];
Visible:=False;
TabStop:=False;
fCaption:='';
fTabVisible:=True;
fImgIndex:=-1;
fClose:=nil;
fShow:=nil;
fHide:=nil;
fTabControl:=nil;
Img:=TImage.Create(nil);
Img.Parent:=Self;
Img.AutoSize:=True;
Img.Align:=alClient;
Img.SendToBack;
end;

destructor TMCoPXPTabSheet.Destroy;
begin
Img.Free;
inherited Destroy;
end;

procedure TMCoPXPTabSheet.SetCaption(V:String);
begin
fCaption:=V;
FindTabControl;
if fTabControl<>nil then fTabControl.Ref;
end;

procedure TMCoPXPTabSheet.SetImgIndex(V:TImageIndex);
begin
fImgIndex:=V;
FindTabControl;
if fTabControl<>nil then fTabControl.Ref;
end;

procedure TMCoPXPTabSheet.SetTabVisible(V:Boolean);
begin
if fTabVisible=V then Exit;
fTabVisible:=V;
if (V) and (Assigned(fShow)) then fShow(Self);
if (not V) and (Assigned(fHide)) then fHide(Self);
FindTabControl;
if fTabControl<>nil then
  begin
  if (fTabControl.ActiveTab=Self) and (V=False) then
    fTabControl.RefActiveTab
  else
    fTabControl.Ref;
  end;
end;

function TMCoPXPTabSheet.GetIndex:Integer;
var k:Integer;
begin
Result:=-1;
FindTabControl;
if fTabControl=nil then Exit;
for k:=0 to fTabControl.ControlCount-1 do
  if fTabControl.Controls[k]=Self then
    begin
    Result:=k;
    Break;
    end;
end;

procedure TMCoPXPTabSheet.SetIndex(V:Integer);
var i:Integer;
begin
FindTabControl;
if fTabControl=nil then Exit;
i:=GetIndex;
if V=i then Exit;
if V>=fTabControl.ControlCount then
  raise EMCoPXPTabError.CreateFmt('%d is an invalid PageIndex value.'#13#10'PageIndex must be between 0 and %d',[V,fTabControl.ControlCount-1]);
BringToFront;
for i:=V to fTabControl.ControlCount-2 do
  fTabControl.Controls[V].BringToFront;
fTabControl.Ref;
end;

procedure TMCoPXPTabSheet.ShowTab;
begin
if not fTabVisible then SetTabVisible(True);
end;

procedure TMCoPXPTabSheet.HideTab;
begin
if fTabVisible then SetTabVisible(False);
end;

procedure TMCoPXPTabSheet.FindTabControl;
begin
if csDestroying in ComponentState then Exit;
if fTabControl=nil then
  if Parent<>nil then
    begin
    fTabControl:=Parent as TMCoPXPTabControl;
    Img.Picture.Assign(fTabControl.SheetBack);
    Img.SendToBack;
    end;
end;

function TMCoPXPTabSheet.GetShowIndex:Integer;
var i,k:Integer;
begin
if csDesigning in ComponentState then
  begin
  Result:=GetIndex;
  Exit;
  end;
Result:=-1;
FindTabControl;
if not Assigned(fTabControl) then Exit;
if not fTabVisible then Exit;
i:=0;
for k:=0 to fTabControl.ControlCount-1 do
  begin
  if fTabControl.TabSheet[k]=Self then Break;
  if fTabControl.TabSheet[k].TabVisible then Inc(i);
  end;
Result:=i;
end;

procedure TMCoPXPTabSheet.Notification(AComponent:TComponent;Operation:TOperation);
begin
inherited Notification(AComponent,Operation);
if csDestroying in ComponentState then Exit;
if (Operation=opRemove) and (AComponent=fTabControl) then fTabControl:=nil;
end;

function TMCoPXPTabSheet.GetControlCount:Integer;
begin
Result:=inherited ControlCount;
Dec(Result);
end;

function TMCoPXPTabSheet.GetControl(Index:Integer):TControl;
begin
Img.SendToBack;
Result:=inherited Controls[Index+1];
end;

procedure TMCoPXPTabSheet.Assign(Source:TPersistent);
var s:TMCoPXPTabSheet;
begin
if Source is TMCoPXPTabSheet then
  begin
  s:=Source as TMCoPXPTabSheet;
  Caption:=s.Caption;
  ImageIndex:=s.ImageIndex;
  TabVisible:=s.TabVisible;
  end
else
  inherited Assign(Source);
end;

procedure TMCoPXPTabSheet.AssignTo(Dest:TPersistent);
var d:TMCoPXPTabSheet;
begin
if Dest is TMCoPXPTabSheet then
  begin
  d:=Dest as TMCoPXPTabSheet;
  d.Caption:=Caption;
  d.ImageIndex:=ImageIndex;
  d.TabVisible:=TabVisible;
  end
else
  inherited AssignTo(Dest);
end;

procedure TMCoPXPTabSheet.CMVisibleChanged(var Message:TMessage);
begin
inherited;
FindTabControl;
if not Assigned(fTabControl) then Exit;
if (fTabControl.ActiveTab<>Self) and (Visible) then fTabControl.ActiveTab:=Self;
end;

end.
